import os
import copy
import warnings
import numpy as np
from yggdrasil import backwards
from yggdrasil.metaschema.encoder import encode_json, decode_json
from yggdrasil.metaschema.datatypes import _schema_dir
from yggdrasil.metaschema.datatypes.JSONObjectMetaschemaType import (
    JSONObjectMetaschemaType)


_schema_file = os.path.join(_schema_dir, 'ply.json')
_index_type = ('int', 'uint')
_color_type = ('int', 'uint')
_coord_type = 'float'
_index_fmt = '%d'
_color_fmt = '%d'
_coord_fmt = '%6.4f'
_index_conv = np.int32
_color_conv = np.uint8
_coord_conv = np.float32
_map_ply2py = {'char': 'int8', 'uchar': 'uint8',
               'short': 'int16', 'ushort': 'uint16',
               'int': 'int32', 'uint': 'uint32',
               'float': 'float32', 'double': 'float64'}
_map_py2ply = {v: k for k, v in _map_ply2py.items()}
_default_element_order = ['material', 'vertices', 'faces', 'edges']
_default_property_order = {'vertices': ['x', 'y', 'z', 'red', 'green', 'blue'],
                           'faces': [],
                           'edges': ['vertex1', 'vertex2', 'red', 'green', 'blue']}
# 'materials': ['ambient_red', 'ambient_green',
#               'ambient_blue', 'ambient_coeff',
#               'diffuse_red', 'diffuse_green',
#               'diffuse_blue', 'diffuse_coeff',
#               'specular_red', 'specular_green',
#               'specular_blue', 'specular_coeff',
#               'specular_power']}


def create_schema(overwrite=False):
    r"""Creates a file containing the Ply schema.

    Args:
        overwrite (bool, optional): If True and a file already exists, the
            existing file will be replaced. If False, an error will be raised
            if the file already exists.

    """
    if (not overwrite) and os.path.isfile(_schema_file):
        raise RuntimeError("Schema file already exists.")
    schema = {
        'title': 'ply',
        'description': 'A mapping container for Ply 3D data.',
        'type': 'object',
        'required': ['vertices', 'faces'],
        'definitions': {
            'index': {'type': ('int', 'uint')},
            'color': {'type': ('int', 'uint')},
            'coord': {'type': 'float'},
            'vertex': {
                'description': 'Map describing a single vertex.',
                'type': 'object', 'required': ['x', 'y', 'z'],
                'additionalProperties': False,
                'properties': {'x': {'type': _coord_type},
                               'y': {'type': _coord_type},
                               'z': {'type': _coord_type},
                               'red': {'type': _color_type},
                               'blue': {'type': _color_type},
                               'green': {'type': _color_type}}},
            'face': {
                'description': 'Map describing a single face.',
                'type': 'object', 'required': ['vertex_index'],
                'additionalProperties': False,
                'properties': {
                    'vertex_index': {
                        'type': 'array', 'minItems': 3,
                        'items': {'type': _index_type}}}},
            'edge': {
                'description': 'Vertex indices describing an edge.',
                'type': 'object', 'required': ['vertex1', 'vertex2'],
                'additionalProperties': False,
                'properties': {
                    'vertex1': {'type': _index_type},
                    'vertex2': {'type': _index_type},
                    'red': {'type': _color_type},
                    'green': {'type': _color_type},
                    'blue': {'type': _color_type}}}
            # 'material': {
            #     'description': 'Map of material parameters.',
            #     'type': 'object',
            #     'required': ['ambient_red', 'ambient_green',
            #                  'ambient_blue', 'ambient_coeff',
            #                  'diffuse_red', 'diffuse_green',
            #                  'diffuse_blue', 'diffuse_coeff',
            #                  'specular_red', 'specular_green',
            #                  'specular_blue', 'specular_coeff',
            #                  'specular_power'],
            #     'properties': {'ambient_red': {'type': _color_type},
            #                    'ambient_green': {'type': _color_type},
            #                    'ambient_blue': {'type': _color_type},
            #                    'ambient_coeff': {'type': _coord_type},
            #                    'diffuse_red': {'type': _color_type},
            #                    'diffuse_green': {'type': _color_type},
            #                    'diffuse_blue': {'type': _color_type},
            #                    'diffuse_coeff': {'type': _coord_type},
            #                    'specular_red': {'type': _color_type},
            #                    'specular_green': {'type': _color_type},
            #                    'specular_blue': {'type': _color_type},
            #                    'specular_coeff': {'type': _coord_type},
            #                    'specular_power': {'type': _coord_type}}}},
        },
        'properties': {
            'material': {
                'description': 'Name of the material to use.',
                'type': ['unicode', 'string']},
            # 'materials': {
            #     'description': 'Array of materials.',
            #     'type': 'array', 'items': {'$ref': '#/definitions/material'}},
            'vertices': {
                'description': 'Array of vertices.',
                'type': 'array', 'items': {'$ref': '#/definitions/vertex'}},
            'edges': {
                'description': 'Array of edges.',
                'type': 'array', 'items': {'$ref': '#/definitions/edge'}},
            'faces': {
                'description': 'Array of faces.',
                'type': 'array', 'items': {'$ref': '#/definitions/face'}}},
        'dependencies': {
            'edges': ['vertices'],
            'faces': ['vertices']}}
    with open(_schema_file, 'w') as fd:
        encode_json(schema, fd, indent='\t')


def get_schema():
    r"""Return the Ply schema, initializing it if necessary.

    Returns:
        dict: Ply schema.
    
    """
    if not os.path.isfile(_schema_file):
        create_schema()
    with open(_schema_file, 'r') as fd:
        out = decode_json(fd)
    return out


if not os.path.isfile(_schema_file):  # pragma: debug
    create_schema()
    

def translate_ply2fmt(type_ply):
    r"""Get the corresponding type string for a Ply type string.

    Args:
        type_ply (str): Ply type string.
    
    Returns:
        str: C-style format string.
    
    """
    if type_ply in ['char', 'uchar', 'short', 'ushort', 'int', 'uint']:
        out = '%d '
    elif type_ply in ['float', 'double']:
        out = '%6.4f '
    else:
        raise ValueError("Could not get format string for type '%s'" % type_ply)
    return out


def translate_ply2py(type_ply):
    r"""Get the corresponding Python type for the Ply type string.

    Args:
        type_ply (str): Ply type string.

    Returns:
        type: Python type.

    Raises:
        ValueError: If the type string does not have a match.

    """
    if type_ply not in _map_ply2py:
        raise ValueError("Could not find type for ply type string '%s'." % type_ply)
    return np.dtype(_map_ply2py[type_ply]).type


def translate_py2ply(py_obj):
    r"""Get the correpsonding Ply type string for the provided Python object.

    Args:
        py_obj (object): Python object.

    Returns:
        str: Ply type string.

    """
    type_py = np.array([py_obj]).dtype
    type_np = np.dtype(type_py).name
    if type_np == 'int64':
        warnings.warn("Ply dosn't support long. Precision will be lost.")
        return _map_py2ply['int32']
    if type_np not in _map_py2ply:
        raise ValueError("Could not find ply type string for numpy type '%s'." % type_np)
    return _map_py2ply[type_np]


def singular2plural(e_sing):
    r"""Get the plural version of a singular element name. If the singular
    version ends with the suffix 'ex' it is replaced with the plural suffix
    'ices'. Otherwise, an 's' is appended to the singular name to make it
    plural.

    Args:
        e_sing (str): Singular version of an element name.

    Returns:
        str: Plural version of the singular element name e_sing.

    """
    if e_sing.endswith('ex'):
        e_plur = e_sing.rsplit('ex', 1)[0]
        e_plur += 'ices'
    else:
        e_plur = e_sing + 's'
    return e_plur


def plural2singular(e_plur):
    r"""Get the singular version of a plural element name. If the plural version
    ends with the suffix 'ices', it is replaced with the singular suffix 'ex'.
    If the plural version ends with an 's', it is removed.

    Args:
        e_plur (str): Plural version of an element name.

    Returns:
        str: Singular version of the plural element name e_plur.

    Raises:
        ValueError: If a singular version cannot be determined.

    """
    if e_plur.endswith('ices'):
        e_sing = e_plur.rsplit('ices', 1)[0]
        e_sing += 'ex'
    elif e_plur.endswith('s'):
        e_sing = e_plur[:-1]
    else:
        raise ValueError("Cannot determine singular version of '%s'." % e_plur)
    return e_sing


class PlyDict(dict):
    r"""Enhanced dictionary class for storing Ply information."""

    def __init__(self, *args, **kwargs):
        super(PlyDict, self).__init__(*args, **kwargs)
        self.setdefault('vertices', [])
        self.setdefault('faces', [])
        self._type_class.validate(self)

    @classmethod
    def from_dict(cls, in_dict):
        r"""Get a version of the object from a dictionary."""
        out = cls(**in_dict)
        return out

    def as_dict(self):
        r"""Get a version of the object as a pure dictionary."""
        out = dict(**self)
        return out
    
    def count_elements(self, element_name):
        r"""Get the count of a certain element in the dictionary.

        Args:
            element_name (str): Name of the element to count.

        Returns:
            int: Number of the provided element.

        """
        if element_name in self:
            return len(self[element_name])
        elif singular2plural(element_name) in self:
            return len(self[singular2plural(element_name)])
        else:
            raise ValueError("'%s' is not a valid property." % element_name)

    @property
    def nvert(self):
        r"""int: Number of vertices."""
        return self.count_elements('vertices')

    @property
    def nface(self):
        r"""int: Number of faces."""
        return self.count_elements('faces')

    @property
    def bounds(self):
        r"""tuple: Mins/maxs of vertices in each dimension."""
        mins = np.empty(3, 'float64')
        maxs = np.empty(3, 'float64')
        for i, x in enumerate('xyz'):
            mins[i] = min([v[x] for v in self['vertices']])
            maxs[i] = max([v[x] for v in self['vertices']])
        return mins, maxs

    @property
    def mesh(self):
        r"""list: Vertices for each face in the structure."""
        mesh = []
        for i in range(self.count_elements('faces')):
            imesh = []
            for f in self['faces']:
                for v in f['vertex_index']:
                    imesh += [self['vertices'][v][k] for k in ['x', 'y', 'z']]
            mesh.append(imesh)
        return mesh

    @classmethod
    def from_shape(cls, shape, d, conversion=1.0, _as_obj=False):  # pragma: lpy
        r"""Create a ply dictionary from a PlantGL shape and descritizer.

        Args:
            scene (openalea.plantgl.scene): Scene that should be descritized.
            d (openalea.plantgl.descritizer): Descritizer.
            conversion (float, optional): Conversion factor that should be
                applied to the vertex positions. Defaults to 1.0.

        """
        out = None
        d.process(shape)
        if d.result is not None:
            out = cls()
            # Vertices
            for p in d.result.pointList:
                new_vert = {}
                for k in ['x', 'y', 'z']:
                    new_vert[k] = conversion * getattr(p, k)
                out['vertices'].append(new_vert)
            # Colors
            if d.result.colorPerVertex and d.result.colorList:
                if d.result.isColorIndexListToDefault():
                    for i, c in enumerate(d.result.colorList):
                        for k in ['red', 'green', 'blue']:
                            out['vertices'][i][k] = getattr(c, k)
                else:  # pragma: debug
                    raise Exception("Indexed vertex colors not supported.")
            # elif not shape.appearance.isAmbientToDefault():
            #     c = shape.appearance.ambient
            #     for k in ['red', 'green', 'blue']:
            #         for v in out['vertices']:
            #             v[k] = getattr(c, k)
            # Material
            if (shape.appearance.name != shape.appearance.DEFAULT_MATERIAL.name):
                out['material'] = shape.appearance.name
            # Faces
            if _as_obj:
                for i3 in d.result.indexList:
                    out['faces'].append([{'vertex_index': i3[j]}
                                         for j in range(len(i3))])
            else:
                for i3 in d.result.indexList:
                    out['faces'].append({'vertex_index': [i3[j] for j in
                                                          range(len(i3))]})
        return out

    @classmethod
    def from_scene(cls, scene, d=None, conversion=1.0):  # pragma: lpy
        r"""Create a ply dictionary from a PlantGL scene and descritizer.

        Args:
            scene (openalea.plantgl.scene): Scene that should be descritized.
            d (openalea.plantgl.descritizer, optional): Descritizer. Defaults
                to openalea.plantgl.all.Tesselator.
            conversion (float, optional): Conversion factor that should be
                applied to the vertex positions. Defaults to 1.0.

        """
        if d is None:
            from openalea.plantgl.all import Tesselator
            d = Tesselator()
        out = cls()
        for k, shapes in scene.todict().items():
            for shape in shapes:
                d.clear()
                iply = cls.from_shape(shape, d, conversion=conversion)
                if iply is not None:
                    out.append(iply)
                d.clear()
        return out

    def to_scene(self, conversion=1.0, name=None):  # pragma: lpy
        r"""Create a PlantGL scene from a Ply dictionary.

        Args:
            conversion (float, optional): Conversion factor that should be
                applied to the vertices. Defaults to 1.0.
            name (str, optional): Name that should be given to the created
                PlantGL symbol. Defaults to None and is ignored.

        Returns:
        

        """
        import openalea.plantgl.all as pgl
        smb_class, args, kwargs = self.to_geom_args(conversion=conversion,
                                                    name=name)
        smb = smb_class(*args, **kwargs)
        if name is not None:
            smb.setName(name)
        if self.get('material', None) is not None:
            mat = pgl.Material(self['material'])
            shp = pgl.Shape(smb, mat)
        else:
            shp = pgl.Shape(smb)
        if name is not None:
            shp.setName(name)
        scn = pgl.Scene([shp])
        return scn

    def to_geom_args(self, conversion=1.0, name=None, _as_obj=False):  # pragma: lpy
        r"""Get arguments for creating a PlantGL geometry.

        Args:
            conversion (float, optional): Conversion factor that should be
                applied to the vertices. Defaults to 1.0.
            name (str, optional): Name that should be given to the created
                PlantGL symbol. Defaults to None and is ignored.

        Returns:
            tuple: Class, arguments and keyword arguments for PlantGL geometry.

        """
        import openalea.plantgl.all as pgl
        kwargs = dict()
        # Add vertices
        obj_points = []
        obj_colors = []
        for v in self['vertices']:
            xarr = conversion * np.array([v[k] for k in ['x', 'y', 'z']])
            obj_points.append(pgl.Vector3(np.float64(xarr[0]),
                                          np.float64(xarr[1]),
                                          np.float64(xarr[2])))
            c = [v.get(k, None) for k in ['red', 'green', 'blue']]
            if None not in c:
                cast_type = int
                obj_colors.append(pgl.Color4(cast_type(c[0]),
                                             cast_type(c[1]),
                                             cast_type(c[2]),
                                             cast_type(1)))
        points = pgl.Point3Array(obj_points)
        if obj_colors:
            colors = pgl.Color4Array(obj_colors)
            kwargs['colorList'] = colors
            kwargs['colorPerVertex'] = True
        # Add indices
        obj_indices = []
        index_class = pgl.Index
        array_class = pgl.IndexArray
        smb_class = pgl.FaceSet
        # index_class = pgl.Index3
        # array_class = pgl.Index3Array
        # smb_class = pgl.TriangleSet
        for f in self['faces']:
            if _as_obj:
                f_int = [int(_f['vertex_index']) for _f in f]
            else:
                f_int = [int(_f) for _f in f['vertex_index']]
            obj_indices.append(index_class(*f_int))
        indices = array_class(obj_indices)
        args = (points, indices)
        return smb_class, args, kwargs

    def append(self, solf):
        r"""Append new ply information to this dictionary.

        Args:
            solf (PlyDict): Another ply to append to this one.

        """
        nvert = self.count_elements('vertices')
        # Vertex fields
        self['vertices'] += solf['vertices']
        # Face fields
        for f in solf['faces']:
            self['faces'].append({'vertex_index': [v + nvert for v in
                                                   f['vertex_index']]})
        # Edge fields
        if 'edges' in solf:
            if 'edges' not in self:
                self['edges'] = []
            for e in solf['edges']:
                iedge = {'vertex1': e['vertex1'] + nvert,
                         'vertex2': e['vertex2'] + nvert}
                for k in ['red', 'green', 'blue']:
                    if k in e:
                        iedge[k] = e[k]
                self['edges'].append(iedge)

    def merge(self, ply_list, no_copy=False):
        r"""Merge a list of ply dictionaries.

        Args:
            ply_list (list): Ply dictionaries.
            no_copy (bool, optional): If True, the current dictionary will be
                updated, otherwise a copy will be returned with the update.
                Defaults to False.

        Returns:
            dict: Merged ply dictionary.

        """
        if not isinstance(ply_list, list):
            ply_list = [ply_list]
        # Merge fields
        if no_copy:
            out = self
        else:
            out = copy.deepcopy(self)
        for x in ply_list:
            out.append(x)
        return out

    def apply_scalar_map(self, scalar_arr, color_map=None,
                         vmin=None, vmax=None, scaling='linear',
                         scale_by_area=False, no_copy=False, _as_obj=False):
        r"""Set the color of faces in a 3D object based on a scalar map.
        This creates a copy unless no_copy is True.

        Args:
            scalar_arr (arr): Scalar values that should be mapped to colors
                for each face.
            color_map (str, optional): The name of the color map that should
                be used. Defaults to 'plasma'.
            vmin (float, optional): Value that should map to the minimum of the
                colormap. Defaults to min(scalar_arr).
            vmax (float, optional): Value that should map to the maximum of the
                colormap. Defaults to max(scalar_arr).
            scaling (str, optional): Scaling that should be used to map the scalar
                array onto the colormap. Defaults to 'linear'.
            scale_by_area (bool, optional): If True, the elements of the scalar
                array will be multiplied by the area of the corresponding face.
                If True, vmin and vmax should be in terms of the scaled array.
                Defaults to False.
            no_copy (bool, optional): If True, the returned object will not be a
                copy. Defaults to False.

        Returns:
            dict: Ply with updated vertex colors.

        """
        from matplotlib import cm
        from matplotlib import colors as mpl_colors
        # Scale by area
        if scale_by_area:
            scalar_arr = copy.deepcopy(scalar_arr)
            for i, f in enumerate(self['faces']):
                if _as_obj:
                    fv = [_f['vertex_index'] for _f in f]
                else:
                    fv = f['vertex_index']
                if len(fv) > 3:
                    raise NotImplementedError("Area calc not implemented "
                                              + "for faces above triangle.")
                v0 = np.array([self['vertices'][fv[0]][k] for k in 'xyz'])
                v1 = np.array([self['vertices'][fv[1]][k] for k in 'xyz'])
                v2 = np.array([self['vertices'][fv[2]][k] for k in 'xyz'])
                a = np.sqrt(np.sum((v0 - v1)**2))
                b = np.sqrt(np.sum((v1 - v2)**2))
                c = np.sqrt(np.sum((v2 - v0)**2))
                s = (a + b + c) / 2.0
                area = np.sqrt(s * (s - a) * (s - b) * (s - c))
                scalar_arr[i] = area * scalar_arr[i]
        # Map vertices onto faces
        vertex_scalar = [[] for x in self['vertices']]
        if _as_obj:
            for i in range(len(self['faces'])):
                for v in self['faces'][i]:
                    vertex_scalar[v['vertex_index']].append(scalar_arr[i])
        else:
            for i in range(len(self['faces'])):
                for v in self['faces'][i]['vertex_index']:
                    vertex_scalar[v].append(scalar_arr[i])
        for i in range(len(vertex_scalar)):
            if len(vertex_scalar[i]) == 0:
                vertex_scalar[i] = 0
            else:
                vertex_scalar[i] = np.mean(vertex_scalar[i])
        vertex_scalar = np.array(vertex_scalar)
        if scaling == 'log':
            vertex_scalar = np.ma.MaskedArray(vertex_scalar, vertex_scalar <= 0)
        # Get color scaling
        if color_map is None:
            # color_map = 'summer'
            color_map = 'plasma'
        if vmin is None:
            vmin = vertex_scalar.min()
        if vmax is None:
            vmax = vertex_scalar.max()
        cmap = cm.get_cmap(color_map)
        if scaling == 'log':
            norm = mpl_colors.LogNorm(vmin=vmin, vmax=vmax)
        elif scaling == 'linear':
            norm = mpl_colors.Normalize(vmin=vmin, vmax=vmax)
        else:  # pragma: debug
            raise Exception("Scaling must be 'linear' or 'log'.")
        m = cm.ScalarMappable(norm=norm, cmap=cmap)
        # Scale colors
        vertex_colors = (255 * m.to_rgba(vertex_scalar)).astype('int')[:, :3].tolist()
        if no_copy:
            out = self
        else:
            out = copy.deepcopy(self)
        for i, c in enumerate(vertex_colors):
            for j, k in enumerate(['red', 'green', 'blue']):
                out['vertices'][i][k] = c[j]
        return out


def get_key_order(all_keys, default_order):
    r"""Determine the order of keys based on the keys and default order. Keys
    are added first in the default order and then alphabetically.

    Args:
        all_keys (list): List of all keys that should be present in the returned
            list.
        default_order (list): Default order for keys that may or may not be in
            all_keys.

    Returns:
        list: Key order.

    """

    def sort_key(x):
        if x in default_order:
            return str(default_order.index(x))
        else:
            return x

    out = sorted(all_keys, key=sort_key)
    return out

   
# The base class could be anything since it is discarded during registration,
# but is set to JSONObjectMetaschemaType here for transparancy since this is
# what the base class is determined to be on loading the schema
class PlyMetaschemaType(JSONObjectMetaschemaType):
    r"""Ply 3D structure map."""

    _empty_msg = {'vertices': [], 'faces': []}
    python_types = (dict, PlyDict)
    schema_file = _schema_file
    _replaces_existing = False

    @classmethod
    def encode_data(cls, obj, typedef, element_order=None, property_order=None,
                    default_rgb=[0, 0, 0], comments=[], newline='\n',
                    plyformat='ascii 1.0'):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.
            element_order (list, optional): Order that elements should be written
                to the file. If not provided, the order is determined based on
                typical ply files with remaining elements output in sorted order.
            property_order (dict, optional): Dictionary of property order for
                each element determining the order that they properties should
                be written to the file. If not provided, the orders are determined
                based on typical ply files with remaining elements output in sorted
                order.
            default_rgb (list, optional): Default color in RGB that should be
                used for missing colors. Defaults to [0, 0, 0].
            comments (list, optional): List of comments that should be included in
                the file header. Defaults to lines describing the automated origin
                of the file.
            newline (str, optional): String that should be used to delineated end
                of lines. Defaults to '\n'.
            plyformat (str, optional): String describing the ply format and version.
                Defaults to 'ascii 1.0'.

        Returns:
            bytes, str: Serialized message.

        """
        # Add comments to identify generated files
        default_comments = ['author ygg_auto', 'File generated by yggdrasil']
        for c in default_comments:
            if c not in comments:
                comments.append(c)
        # Default order to allow user definited elements
        if element_order is None:
            element_order = get_key_order(obj.keys(), _default_element_order)
        if property_order is None:
            property_order = {}
            for e in element_order:
                if e == 'material':
                    continue
                assert(isinstance(obj[e], (list, tuple)))
                if len(obj[e]) == 0:
                    continue
                property_order[e] = get_key_order(obj[e][0].keys(),
                                                  _default_property_order.get(e, []))
        # Get information needed
        size_map = {}
        type_map = {}
        for e in element_order:
            if e == 'material':
                continue
            assert(isinstance(obj[e], (list, tuple)))
            type_map[e] = {}
            size_map[e] = len(obj[e])
            if size_map[e] == 0:
                continue
            for p in property_order[e]:
                if isinstance(obj[e][0][p], list):
                    subtype = translate_py2ply(obj[e][0][p][0])
                    type_map[e][p] = 'list uchar %s' % subtype
                else:
                    type_map[e][p] = translate_py2ply(obj[e][0][p])
        # Encode header
        header = ['ply', 'format %s' % plyformat]
        header += ['comment ' + c for c in comments]
        for e in element_order:
            if e == 'material':
                header += ['comment material: %s' % obj[e]]
            else:
                e_sing = plural2singular(e)
                header.append('element %s %d' % (e_sing, size_map[e]))
                if size_map[e] > 0:
                    for p in property_order[e]:
                        header.append('property %s %s' % (type_map[e][p], p))
        header.append('end_header')
        # Encode body
        body = []
        for e in element_order:
            if (e not in obj) or (e == 'material'):
                continue
            for x in obj[e]:
                iline = ''
                for p in property_order[e]:
                    if type_map[e][p].startswith('list'):
                        vars = type_map[e][p].split()
                        ientry = x[p]
                        ifmt = translate_ply2fmt(vars[2])
                        iline += translate_ply2fmt(vars[1]) % len(ientry)
                        for ix in ientry:
                            iline += ifmt % ix
                    else:
                        # if p not in x:
                        #     if p in ['red', 'green', 'blue']:
                        #         rgb_index = ['red', 'green', 'blue'].index(p)
                        #         x[p] = default_rgb[rgb_index]
                        iline += translate_ply2fmt(type_map[e][p]) % x[p]
                body.append(iline.strip())  # Ensure trailing spaces are removed
        return newline.join(header + body) + newline
        
    @classmethod
    def encode_data_readable(cls, obj, typedef):
        r"""Encode an object's data in a readable format.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        return cls.encode_data(obj, typedef)
    
    @classmethod
    def decode_data(cls, msg, typedef):
        r"""Decode an object.

        Args:
            msg (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        lines = backwards.as_str(msg).splitlines()
        metadata = {'comments': [], 'element_order': [], 'property_order': {}}
        if lines[0] != 'ply':
            raise ValueError("The first line must be 'ply'")
        # Parse header
        e = None
        p = None
        type_map = {}
        size_map = {}
        obj = {}
        for i, line in enumerate(lines):
            if line.startswith('format'):
                metadata['plyformat'] = line.split(None, 1)[-1]
            elif line.startswith('comment'):
                out = line.split(None, 1)[-1]
                if out.startswith('material:'):
                    metadata['element_order'].append('material')
                    obj['material'] = out.split(None, 1)[-1]
                metadata['comments'].append(out)
            elif line.startswith('element'):
                vars = line.split()
                e_sing = vars[1]
                e = singular2plural(e_sing)
                size_map[e] = int(float(vars[2]))
                type_map[e] = {}
                metadata['element_order'].append(e)
                metadata['property_order'][e] = []
                obj[e] = []
            elif line.startswith('property'):
                vars = line.split()
                p = vars[-1]
                type_map[e][p] = ' '.join(vars[1:-1])
                metadata['property_order'][e].append(p)
            elif 'end_header' in line:
                headline = i + 1
                break
        # Parse body
        i = headline
        for e in metadata['element_order']:
            if e == 'material':
                continue
            for ie in range(size_map[e]):
                vars = lines[i].split()
                iv = 0
                new = {}
                for p in metadata['property_order'][e]:
                    if type_map[e][p].startswith('list'):
                        type_vars = type_map[e][p].split()
                        count_type = translate_ply2py(type_vars[1])
                        plist_type = translate_ply2py(type_vars[2])
                        count = count_type(vars[iv])
                        plist = []
                        iv += 1
                        for ip in range(count):
                            plist.append(plist_type(vars[iv]))
                            iv += 1
                        new[p] = plist
                    else:
                        prop_type = translate_ply2py(type_map[e][p])
                        new[p] = prop_type(vars[iv])
                        iv += 1
                assert(iv == len(vars))
                obj[e].append(new)
                i += 1
        # Check that all properties filled in
        for e in metadata['element_order']:
            if e not in metadata['property_order']:
                continue
            for p in metadata['property_order'][e]:
                assert(len(obj[e]) == size_map[e])
        # Return
        return PlyDict(obj)

    @classmethod
    def updated_fixed_properties(cls, obj):
        r"""Get a version of the fixed properties schema that includes information
        from the object.

        Args:
            obj (object): Object to use to put constraints on the fixed properties
                schema.

        Returns:
            dict: Fixed properties schema with object dependent constraints.

        """
        out = super(PlyMetaschemaType, cls).updated_fixed_properties(obj)
        # Constrain indices on number of elements they refer to
        if isinstance(obj, dict) and ('vertices' in obj):
            nvert = len(obj['vertices'])
            out['definitions']['face']['properties']['vertex_index']['items'][
                'maximum'] = nvert - 1
            for k in ['vertex1', 'vertex2']:
                out['definitions']['edge']['properties'][k][
                    'maximum'] = nvert - 1
        return out


PlyDict._type_class = PlyMetaschemaType



"""Module for funneling messages from one comm to another."""
import os
import numpy as np
import threading
from yggdrasil import backwards
from yggdrasil.communication import new_comm
from yggdrasil.drivers.Driver import Driver
from yggdrasil.components import import_component
from yggdrasil.schema import get_schema


def _translate_list2element(arr):
    if isinstance(arr, (list, tuple)):
        arr = arr[0]
    return arr


class ConnectionDriver(Driver):
    r"""Class that continuously passes messages from one comm to another.

    Args:
        name (str): Name that should be used to set names of input/output comms.
        icomm_kws (dict, optional): Keyword arguments for the input communicator.
        ocomm_kws (dict, optional): Keyword arguments for the output communicator.
        translator (str, func, optional): Function or string specifying function
            that should be used to translate messages from the input communicator
            before passing them to the output communicator. If a string, the
            format should be "<package.module>:<function>" so that <function>
            can be imported from <package>. Defaults to None and messages are
            passed directly. This can also be a list of functions/strings that
            will be called on the messages in the order they are provided.
        timeout_send_1st (float, optional): Time in seconds that should be
            waited before giving up on the first send. Defaults to self.timeout.
        single_use (bool, optional): If True, the driver will be stopped after
            one loop. Defaults to False.
        onexit (str, optional): Class method that should be called when a
            model that the connection interacts with exits, but before the
            connection driver is shut down. Defaults to None.
        **kwargs: Additonal keyword arguments are passed to the parent class.

    Attributes:
        icomm_kws (dict): Keyword arguments for the input communicator.
        ocomm_kws (dict): Keyword arguments for the output communicator.
        icomm (CommBase): Input communicator.
        ocomm (CommBase): Output communicator.
        nrecv (int): Number of messages received.
        nproc (int): Number of messages processed.
        nsent (int): Number of messages sent.
        nskip (int): Number of messages skipped.
        state (str): Descriptor of last action taken.
        translator (func): Function that will be used to translate messages from
            the input communicator before passing them to the output communicator.
        timeout_send_1st (float): Time in seconds that should be waited before
            giving up on the first send.
        single_use (bool): If True, the driver will be stopped after one
            loop.
        onexit (str): Class method that should be called when the corresponding
            model exits, but before the driver is shut down.

    """

    _connection_type = 'default'
    _icomm_type = 'DefaultComm'
    _ocomm_type = 'DefaultComm'
    _direction = 'any'
    _schema_type = 'connection'
    _schema_subtype_key = 'connection_type'
    _schema_subtype_description = ('Connection between one or more comms/files '
                                   'and one or more comms/files.')
    _schema_subtype_default = 'default'
    _schema_required = ['inputs', 'outputs']
    _schema_properties = {
        'connection_type': {'type': 'string'},  # 'default': 'default'},
        'inputs': {'type': 'array', 'minItems': 1,
                   'items': {'anyOf': [{'$ref': '#/definitions/comm'},
                                       {'$ref': '#/definitions/file'}]},
                   'description': ('One or more name(s) of model output(s) '
                                   'and/or new comm/file objects that the '
                                   'connection should receive messages from.')},
        'outputs': {'type': 'array', 'minItems': 1,
                    'items': {'anyOf': [{'$ref': '#/definitions/comm'},
                                        {'$ref': '#/definitions/file'}]},
                    'description': ('One or more name(s) of model input(s) '
                                    'and/or new comm/file objects that the '
                                    'connection should send messages to.')},
        'translator': {'type': 'array', 'items': {'type': 'function'}},
        'onexit': {'type': 'string'}}
    _schema_excluded_from_class_validation = ['inputs', 'outputs']

    @property
    def _is_input(self):
        r"""bool: True if the connection is providing input to a model."""
        return (self._direction == 'input')

    @property
    def _is_output(self):
        r"""bool: True if the connection is retreiving output from a model."""
        return (self._direction == 'output')

    def __init__(self, name, translator=None, single_use=False, onexit=None, **kwargs):
        super(ConnectionDriver, self).__init__(name, **kwargs)
        # Translator
        if translator is None:
            translator = []
        elif not isinstance(translator, list):
            translator = [translator]
        self.translator = []
        for t in translator:
            if not hasattr(t, '__call__'):
                raise ValueError("Translator %s not callable." % t)
            self.translator.append(t)
        if (onexit is not None) and (not hasattr(self, onexit)):
            raise ValueError("onexit '%s' is not a class method." % onexit)
        self.onexit = onexit
        # Attributes
        self._eof_sent = False
        self.single_use = single_use
        self._first_send_done = False
        self._comm_opened = threading.Event()
        self._comm_closed = False
        self._used = False
        self._skip_after_loop = False
        self._model_exited = False
        self.nrecv = 0
        self.nproc = 0
        self.nsent = 0
        self.nskip = 0
        self.state = 'started'
        self.close_state = ''
        # Add comms and print debug info
        self._init_comms(name, **kwargs)
        # self.debug('    env: %s', str(self.env))
        self.debug(('\n' + 80 * '=' + '\n'
                    + 'class = %s\n'
                    + '    input: name = %s, address = %s\n'
                    + '    output: name = %s, address = %s\n'
                    + (80 * '=')), self.__class__,
                   self.icomm.name, self.icomm.address,
                   self.ocomm.name, self.ocomm.address)

    def _init_single_comm(self, name, io, comm_kws, **kwargs):
        r"""Parse keyword arguments for input/output comm."""
        self.debug("Creating %s comm", io)
        s = get_schema()
        if comm_kws is None:
            comm_kws = dict()
        if io == 'input':
            direction = 'recv'
            comm_type = self._icomm_type
            touches_model = self._is_output
            attr_comm = 'icomm'
            comm_kws['close_on_eof_recv'] = False
        else:
            direction = 'send'
            comm_type = self._ocomm_type
            touches_model = self._is_input
            attr_comm = 'ocomm'
        comm_kws['direction'] = direction
        comm_kws['dont_open'] = True
        comm_kws['reverse_names'] = True
        comm_kws.setdefault('comm', {'comm': comm_type})
        assert(name == self.name)
        comm_kws.setdefault('name', name)
        if not isinstance(comm_kws['comm'], list):
            comm_kws['comm'] = [comm_kws['comm']]
        for i, x in enumerate(comm_kws['comm']):
            if x is None:
                comm_kws['comm'][i] = dict()
            elif not isinstance(x, dict):
                comm_kws['comm'][i] = dict(comm=x)
            comm_kws['comm'][i].setdefault('comm', comm_type)
        any_files = False
        all_files = True
        if not touches_model:
            comm_kws['no_suffix'] = True
            ikws = []
            for x in comm_kws['comm']:
                icomm_cls = import_component('comm', x['comm'])
                if icomm_cls.is_file:
                    any_files = True
                    ikws += s['file'].get_subtype_properties(icomm_cls._filetype)
                else:
                    all_files = False
                    ikws += s['comm'].get_subtype_properties(icomm_cls._commtype)
            ikws = list(set(ikws))
            for k in ikws:
                if (k not in comm_kws) and (k in kwargs):
                    comm_kws[k] = kwargs.pop(k)
            if ('comm_env' in kwargs) and ('comm_env' not in comm_kws):
                comm_kws['env'] = kwargs.pop('comm_env')
        if any_files and (io == 'input'):
            kwargs.setdefault('timeout_send_1st', 60)
        self.debug('%s comm_kws:\n%s', attr_comm, self.pprint(comm_kws, 1))
        setattr(self, attr_comm, new_comm(comm_kws.pop('name'), **comm_kws))
        setattr(self, '%s_kws' % attr_comm, comm_kws)
        if touches_model:
            self.env.update(getattr(self, attr_comm).opp_comms)
        elif not all_files:
            self.comm_env.update(getattr(self, attr_comm).opp_comms)
        return kwargs

    def _init_comms(self, name, icomm_kws=None, ocomm_kws=None, **kwargs):
        r"""Parse keyword arguments for input/output comms."""
        kwargs = self._init_single_comm(name, 'input', icomm_kws, **kwargs)
        try:
            kwargs = self._init_single_comm(name, 'output', ocomm_kws, **kwargs)
        except BaseException:
            self.icomm.close()
            raise
        # Apply keywords dependent on comms
        self.timeout_send_1st = kwargs.pop('timeout_send_1st', self.timeout)
        self.debug('Final env:\n%s', self.pprint(self.env, 1))
        
    def wait_for_route(self, timeout=None):
        r"""Wait until messages have been routed."""
        T = self.start_timeout(timeout)
        while ((not T.is_out)
               and (self.nrecv != (self.nsent + self.nskip))):  # pragma: debug
            self.sleep()
        self.stop_timeout()
        return (self.nrecv == (self.nsent + self.nskip))

    @property
    def is_valid(self):
        r"""bool: Returns True if the connection is open and the parent class
        is valid."""
        with self.lock:
            return (super(ConnectionDriver, self).is_valid
                    and self.is_comm_open and not (self.single_use and self._used))

    @property
    def is_comm_open(self):
        r"""bool: Returns True if both communicators are open."""
        with self.lock:
            return (self.icomm.is_open and self.ocomm.is_open
                    and not self._comm_closed)

    @property
    def is_comm_closed(self):
        r"""bool: Returns True if both communicators are closed."""
        with self.lock:
            return self.icomm.is_closed and self.ocomm.is_closed

    @property
    def n_msg(self):
        r"""int: Number of messages waiting in input communicator."""
        with self.lock:
            return self.icomm.n_msg_recv

    def open_comm(self):
        r"""Open the communicators."""
        self.debug('')
        with self.lock:
            if self._comm_closed:
                self.debug('Aborted as comm closed')
                return
            try:
                self.icomm.open()
                self.ocomm.open()
            except BaseException:
                self.close_comm()
                raise
            self._comm_opened.set()
        self.debug('Returning')

    def close_comm(self):
        r"""Close the communicators."""
        self.debug('')
        with self.lock:
            self._comm_closed = True
            self._skip_after_loop = True
            # Capture errors for both comms
            ie = None
            oe = None
            try:
                if getattr(self, 'icomm', None) is not None:
                    self.icomm.close()
            except BaseException as e:
                ie = e
            try:
                if getattr(self, 'ocomm', None) is not None:
                    self.ocomm.close()
            except BaseException as e:
                oe = e
            if ie:
                raise ie
            if oe:
                raise oe
        self.debug('Returning')

    def start(self):
        r"""Open connection before running."""
        self.open_comm()
        Tout = self.start_timeout()
        while (not self.is_comm_open) and (not Tout.is_out):
            self.sleep()
        self.stop_timeout()
        if not self.is_comm_open:
            raise Exception("Connection never finished opening.")
        super(ConnectionDriver, self).start()

    def graceful_stop(self, timeout=None, **kwargs):
        r"""Stop the driver, first waiting for the input comm to be empty.

        Args:
            timeout (float, optional): Max time that should be waited. Defaults
                to None and is set to attribute timeout.
            **kwargs: Additional keyword arguments are passed to the parent
                class's graceful_stop method.

        """
        self.debug('')
        with self.lock:
            self.set_close_state('stop')
            self._skip_after_loop = True
        self.drain_input(timeout=timeout)
        self.wait_for_route(timeout=timeout)
        self.drain_output(timeout=timeout)
        super(ConnectionDriver, self).graceful_stop()
        self.debug('Returning')

    def on_model_exit(self):
        r"""Drain input and then close it."""
        self.debug('')
        if (self.onexit not in [None, 'on_model_exit', 'pass']):
            self.debug("Calling onexit = '%s'" % self.onexit)
            getattr(self, self.onexit)()
        self.drain_input(timeout=self.timeout)
        self.set_close_state('model exit')
        self.debug('Model exit triggered close')
        if self._is_input:
            with self.lock:
                self.icomm.close()
                self.ocomm.close()
        if self._is_output:
            self.wait_for_route(timeout=self.timeout)
            with self.lock:
                self.icomm.close()
        self.set_break_flag()
        self.wait()
        self.debug('Finished')
        super(ConnectionDriver, self).on_model_exit()

    def do_terminate(self):
        r"""Stop the driver by closing the communicators."""
        self.debug('')
        self.set_close_state('terminate')
        self.close_comm()
        super(ConnectionDriver, self).do_terminate()

    def cleanup(self):
        r"""Ensure that the communicators are closed."""
        self.debug('')
        self.close_comm()
        super(ConnectionDriver, self).cleanup()

    def printStatus(self, beg_msg='', end_msg=''):
        r"""Print information on the status of the ConnectionDriver.

        Arguments:
            beg_msg (str, optional): Additional message to print at beginning.
            end_msg (str, optional): Additional message to print at end.

        """
        msg = beg_msg
        msg += '%-50s' % (self.__module__.split('.')[-1] + '(' + self.name + '): ')
        msg += '%-30s' % ('last action: ' + self.state)
        msg += '%-25s' % ('is_open(%s, %s), ' % (self.icomm.is_open,
                                                 self.ocomm.is_open))
        msg += '%-15s' % (str(self.nrecv) + ' received, ')
        msg += '%-15s' % (str(self.nproc) + ' processed, ')
        msg += '%-15s' % (str(self.nskip) + ' skipped, ')
        msg += '%-15s' % (str(self.nsent) + ' sent, ')
        msg += '%-20s' % (str(self.icomm.n_msg) + ' ready to recv')
        msg += '%-20s' % (str(self.ocomm.n_msg) + ' ready to send')
        with self.lock:
            if self.close_state:
                msg += '%-30s' % ('close state: ' + self.close_state)
        msg += end_msg
        print(msg)

    def confirm_input(self, timeout=None):
        r"""Confirm receipt of messages from input comm."""
        T = self.start_timeout(timeout)
        while not T.is_out:  # pragma: debug
            with self.lock:
                if (not self.icomm.is_open):
                    break
                elif self.icomm.is_confirmed_recv:
                    break
            self.sleep(10 * self.sleeptime)
        self.stop_timeout()

    def confirm_output(self, timeout=None):
        r"""Confirm receipt of messages from output comm."""
        T = self.start_timeout(timeout)
        while not T.is_out:  # pragma: debug
            with self.lock:
                if (not self.ocomm.is_open):
                    break
                elif self.ocomm.is_confirmed_send:
                    break
            self.sleep(10 * self.sleeptime)
        self.stop_timeout()

    def drain_input(self, timeout=None):
        r"""Drain messages from input comm."""
        T = self.start_timeout(timeout)
        while not T.is_out:
            with self.lock:
                if (not self.icomm.is_open):
                    break
                elif ((self.icomm.n_msg_recv_drain == 0)
                      and self.icomm.is_confirmed_recv):
                    break
            self.sleep()
        self.stop_timeout()

    def drain_output(self, timeout=None):
        r"""Drain messages from output comm."""
        T = self.start_timeout(timeout)
        while not T.is_out:
            with self.lock:
                if (not self.ocomm.is_open):  # pragma: no cover
                    break
                elif ((self.ocomm.n_msg_send_drain == 0)
                      and self.ocomm.is_confirmed_send):
                    break
            self.sleep()
        self.stop_timeout()

    def before_loop(self):
        r"""Actions to perform prior to sending messages."""
        self.state = 'before loop'
        try:
            self.open_comm()
            self.sleep()  # Help ensure senders/receivers connected before messages
            self.debug('Running in %s, is_valid = %s', os.getcwd(), str(self.is_valid))
            assert(self.is_valid)
        except BaseException:  # pragma: debug
            self.printStatus()
            self.exception('Could not prep for loop (is_open = (%s, %s)).' % (
                self.icomm.is_open, self.ocomm.is_open))
            self.close_comm()
            self.set_break_flag()

    def after_loop(self):
        r"""Actions to perform after sending messages."""
        self.state = 'after loop'
        self.debug('')
        # Close input comm in case loop did not
        self.confirm_input(timeout=False)
        self.debug('Confirmed input')
        with self.lock:
            self.debug('Acquired lock')
            if self._skip_after_loop:
                self.debug("After loop skipped.")
                return
            self.icomm.close()
        # Send EOF in case the model didn't
        if not self.single_use:
            self.send_eof()
        # Do not close output comm in case model/connection still receiving
        self.debug('Finished')

    def recv_message(self, **kwargs):
        r"""Get a new message to send.

        Args:
            **kwargs: Additional keyword arguments are passed to the appropriate
                recv method.

        Returns:
            str, bool: False if no more messages, message otherwise.

        """
        kwargs.setdefault('timeout', 0)
        with self.lock:
            if self.icomm.is_closed:
                return False
            flag, msg = self.icomm.recv(**kwargs)
        if self.icomm.is_eof(msg):
            return self.on_eof()
        if flag:
            return msg
        else:
            return flag

    def on_eof(self):
        r"""Actions to take when EOF received.

        Returns:
            str, bool: Value that should be returned by recv_message on EOF.

        """
        with self.lock:
            self.debug('EOF received')
            self.state = 'eof'
            self.set_close_state('eof')
            self.set_break_flag()
        self.debug('After EOF')
        return False

    def on_message(self, msg):
        r"""Process a message.

        Args:
            msg (bytes, str): Message to be processed.

        Returns:
            bytes, str: Processed message.

        """
        if (self.ocomm._send_serializer) and self.icomm.serializer.initialized:
            self.update_serializer(msg)
        for t in self.translator:
            msg = t(msg)
        return msg

    def update_serializer(self, msg):
        r"""Update the serializer for the output comm based on input."""
        sinfo = self.icomm.serializer.typedef
        sinfo.update(self.icomm.serializer.serializer_info)
        sinfo.pop('seritype', None)
        self.debug('Before update:\n'
                   + '  icomm:\n    sinfo:\n%s\n    typedef:\n%s\n'
                   + '  ocomm:\n    sinfo:\n%s\n    typedef:\n%s',
                   self.pprint(self.icomm.serializer.serializer_info, 2),
                   self.pprint(self.icomm.serializer.typedef, 2),
                   self.pprint(self.ocomm.serializer.serializer_info, 2),
                   self.pprint(self.ocomm.serializer.typedef, 2))
        self.ocomm.serializer.initialize_serializer(sinfo)
        self.ocomm.serializer.update_serializer(skip_type=True,
                                                **self.icomm._last_header)
        if (((self.icomm.serializer.typedef['type'] == 'array')
             and (self.ocomm.serializer.typedef['type'] != 'array')
             and (len(self.icomm.serializer.typedef['items']) == 1))):
            self.translator.insert(0, _translate_list2element)
        # inter_model = False
        # if self.icomm.is_file:
        #     # Remove the file information and only pass the type definition
        #     typedef_in = self.icomm.serializer.typedef
        #     sinfo = self.icomm.serializer.typedef
        #     sinfo.pop('seritype', None)
        # elif self.ocomm.is_file:
        #     # Maintain the default serializer type for the file
        #     sinfo = self.icomm.serializer.serializer_info
        #     sinfo.pop('seritype')
        #     sinfo.update(self.ocomm.serializer.serializer_info)
        #     sinfo.update(self.icomm.serializer.typedef)
        # else:
        #     # Copy the serializer and prevent the type from being overwritten
        #     # TODO: icomm is probably initialized so the serializer info
        #     # from the output comm won't be used.
        #     sinfo = self.ocomm.serializer.serializer_info
        #     sinfo.pop('seritype', None)
        #     self.ocomm.serializer = self.icomm.serializer
        #     inter_model = True
        # if (not inter_model) and self.ocomm.serializer.initialized:  # pragma: debug
        #     self.ocomm.serializer.update_serializer(**sinfo)
        # else:
        #     self.ocomm.serializer.initialize_serializer(sinfo)
        self.debug('After update:\n'
                   + '  icomm:\n    sinfo:\n%s\n    typedef:\n%s\n'
                   + '  ocomm:\n    sinfo:\n%s\n    typedef:\n%s',
                   self.pprint(self.icomm.serializer.serializer_info, 2),
                   self.pprint(self.icomm.serializer.typedef, 2),
                   self.pprint(self.ocomm.serializer.serializer_info, 2),
                   self.pprint(self.ocomm.serializer.typedef, 2))

    def _send_message(self, *args, **kwargs):
        r"""Send a single message.

        Args:
            *args: Arguments are passed to the output comm send method.
            *kwargs: Keyword arguments are passed to the output comm send method.

        Returns:
            bool: Success or failure of send.

        """
        with self.lock:
            if self.ocomm.is_closed:
                return False
            flag = self.ocomm.send(*args, **kwargs)
            return flag
        
    def _send_1st_message(self, *args, **kwargs):
        r"""Send the first message, trying multiple times.

        Args:
            *args: Arguments are passed to the output comm send method.
            *kwargs: Keyword arguments are passed to the output comm send method.

        Returns:
            bool: Success or failure of send.

        """
        self.ocomm._multiple_first_send = False
        T = self.start_timeout(self.timeout_send_1st)
        flag = self._send_message(*args, **kwargs)
        self.ocomm.suppress_special_debug = True
        if not flag:
            self.debug("1st send failed, will keep trying for %f s in silence.",
                       float(self.timeout_send_1st))
        while ((not T.is_out) and (not flag)
               and self.ocomm.is_open):  # pragma: debug
            flag = self._send_message(*args, **kwargs)
            if not flag:
                self.sleep()
        self.stop_timeout()
        self.ocomm.suppress_special_debug = False
        self._first_send_done = True
        if not flag:
            self.error("1st send failed.")
        else:
            self.debug("1st send succeded")
        return flag

    def send_eof(self):
        r"""Send EOF message.

        Returns:
            bool: Success or failure of send.

        """
        with self.lock:
            if self._eof_sent:  # pragma: debug
                self.debug('Already sent EOF')
                return False
            self._eof_sent = True
        self.debug('Sent EOF')
        return self.send_message(self.ocomm.eof_msg, is_eof=True)

    def send_message(self, *args, **kwargs):
        r"""Send a single message.

        Args:
            *args: Arguments are passed to the output comm send method.
            *kwargs: Keyword arguments are passed to the output comm send method.

        Returns:
            bool: Success or failure of send.

        """
        self.debug('')
        kwargs.pop('is_eof', False)
        with self.lock:
            self._used = True
        if self._first_send_done:
            flag = self._send_message(*args, **kwargs)
        else:
            flag = self._send_1st_message(*args, **kwargs)
        # if self.single_use:
        #     with self.lock:
        #         self.debug('Used')
        #         self.icomm.drain_messages()
        #         self.icomm.close()
        return flag

    def set_close_state(self, state):
        r"""Set the close state if its not already set."""
        out = False
        with self.lock:
            if not self.close_state:
                self.debug("Setting close state to %s", state)
                self.close_state = state
                out = True
        return out

    def wait_close_state(self, state, timeout=None):
        r"""Set the close state after waiting for specified time for the
        close state to be set by another method.

        Args:
            state (str): Close state that should be set after timeout.
            timeout (float, optional): Time that should be waited before
                setting the timeout. Defaults to self.timeout.

        """
        T = self.start_timeout(timeout)
        while (not T.is_out):  # pragma: debug
            with self.lock:
                if self.close_state:
                    break
            self.sleep(2 * self.sleeptime)
        self.stop_timeout()
        self.set_close_state(state)

    def run_loop(self):
        r"""Run the driver. Continue looping over messages until there are not
        any left or the communication channel is closed.
        """
        self.state = 'in loop'
        if not self.is_valid:
            self.debug("Breaking loop")
            self.set_close_state('invalid')
            self.set_break_flag()
            return
        # Receive a message
        self.state = 'receiving'
        msg = self.recv_message()
        if msg is False:
            self.debug('No more messages')
            self.set_break_flag()
            self.set_close_state('receiving')
            return
        if self.icomm.is_empty_recv(msg):
            self.state = 'waiting'
            self.verbose_debug(':run: Waiting for next message.')
            self.sleep()
            return
        self.nrecv += 1
        self.state = 'received'
        if isinstance(msg, backwards.bytes_type):
            self.debug('Received message that is %d bytes from %s.',
                       len(msg), self.icomm.address)
        elif isinstance(msg, np.ndarray):
            self.debug('Received array with shape %s and data type %s from %s',
                       msg.shape, msg.dtype, self.icomm.address)
        else:
            self.debug('Received message of type %s from %s',
                       type(msg), self.icomm.address)
        # Process message
        self.state = 'processing'
        msg = self.on_message(msg)
        if msg is False:  # pragma: debug
            self.error('Could not process message.')
            self.set_break_flag()
            self.set_close_state('processing')
            return
        elif self.ocomm.is_empty_send(msg):
            self.debug('Message skipped.')
            self.nskip += 1
            return
        self.nproc += 1
        self.state = 'processed'
        self.debug('Processed message.')
        # Send a message
        self.state = 'sending'
        ret = self.send_message(msg)
        if ret is False:
            self.error('Could not send message.')
            self.set_break_flag()
            self.set_close_state('sending')
            return
        self.nsent += 1
        self.state = 'sent'
        self.debug('Sent message to %s.', self.ocomm.address)



import uuid
import copy
from yggdrasil.communication.tests import test_CommBase as parent


class TestForkComm(parent.TestCommBase):
    r"""Tests for ForkComm communication class."""

    comm = 'ForkComm'
    attr_list = (copy.deepcopy(parent.TestCommBase.attr_list)
                 + ['comm_list', 'curr_comm_index'])
    ncomm = 2

    @property
    def cleanup_comm_classes(self):
        r"""list: Comm classes that should be cleaned up following the test."""
        return set([self.comm] + [None])

    @property
    def send_inst_kwargs(self):
        r"""dict: Keyword arguments for send instance."""
        out = super(TestForkComm, self).send_inst_kwargs
        out['ncomm'] = self.ncomm
        return out

    def test_error_name(self):
        r"""Test error on missing address."""
        self.assert_raises(RuntimeError, self.import_cls, 'test%s' % uuid.uuid4())

    def test_error_send(self):
        r"""Disabled: Test error on send."""
        pass

    def test_error_recv(self):
        r"""Disabled: Test error on recv."""
        pass

    def test_work_comm(self):
        r"""Disabled: Test creating/removing a work comm."""
        pass

    def do_send_recv(self, *args, **kwargs):
        r"""Generic send/recv of a message."""
        if 'eof' not in kwargs.get('send_meth', 'None'):
            kwargs['n_recv'] = self.ncomm
        super(TestForkComm, self).do_send_recv(*args, **kwargs)

    def test_purge(self, **kwargs):
        r"""Test purging messages from the comm."""
        kwargs['nrecv'] = self.ncomm
        super(TestForkComm, self).test_purge(**kwargs)


class TestForkCommList(TestForkComm):
    r"""Tests for ForkComm communication class with construction from address."""
    @property
    def inst_kwargs(self):
        r"""list: Keyword arguments for tested class."""
        out = super(TestForkComm, self).inst_kwargs
        out['comm'] = 'ForkComm'  # To force test of construction from addresses
        return out



from yggdrasil.tests import YggTestClass, assert_raises


class TestYggTest(YggTestClass):
    r"""Test errors for uninitialized YggTestClass."""

    def create_instance(self):
        r"""Dummy overload to prevent initialization."""
        return None

    def test_description(self):
        r"""Get uninitialized description."""
        self.description_prefix
        self.shortDescription()

    def test_import_cls(self):
        r"""Test import class with mod/cls unset."""
        assert_raises(Exception, getattr, self, 'import_cls')
        self._mod = 'drivers'
        assert_raises(Exception, getattr, self, 'import_cls')

    def test_post_teardown_ref(self):
        r"""Test errors on instance ref post teardown."""
        self.teardown()
        assert_raises(RuntimeError, getattr, self, 'instance')



import sys
import importlib
from yggdrasil import tools
from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver


class PythonModelDriver(InterpretedModelDriver):
    r"""Class for running Python models."""

    _schema_subtype_description = ('Model is written in Python.')
    language = 'python'
    language_ext = '.py'
    default_interpreter = sys.executable
    interface_library = 'yggdrasil.interface.YggInterface'
    # supported_comms = ['ipc', 'zmq', 'rmq']
    supported_comms = tools.get_supported_comm()
    supported_comm_options = {
        'ipc': {'platforms': ['MacOS', 'Linux'],
                'libraries': ['sysv_ipc']},
        'zmq': {'libraries': ['zmq']},
        'rmq': {'libraries': ['pika']}}
    type_map = {
        'int': 'numpy.intX',
        'float': 'numpy.floatX',
        'string': 'str',
        'array': 'list',
        'object': 'dict',
        'boolean': 'bool',
        'null': 'None',
        'uint': 'numpy.uintX',
        'complex': 'numpy.complexX',
        'bytes': 'bytes',
        'unicode': 'str',
        '1darray': 'numpy.ndarray',
        'ndarray': 'numpy.ndarray',
        'ply': 'PlyDict',
        'obj': 'ObjDict',
        'schema': 'dict'}
    function_param = {
        'comment': '#',
        'true': 'True',
        'indent': 4 * ' ',
        'print': 'print(\"{message}\")',
        'error': 'raise Exception("{error_msg}")',
        'block_end': '',
        'if_begin': 'if ({cond}):',
        'for_begin': 'for {iter_var} in range({iter_begin}, {iter_end}):',
        'while_begin': 'while ({cond}):',
        'try_begin': 'try:',
        'try_error_type': 'BaseException',
        'try_except': 'except {error_type} as {error_var}:',
        'assign': '{name} = {value}',
        'exec_begin': 'def main():',
        'exec_suffix': ('if __name__ == "__main__":\n'
                        '    main()')}

    @classmethod
    def is_language_installed(self):
        r"""Determine if this model driver is installed on the current
        machine.

        Returns:
            bool: Truth of if this model driver can be run on the current
                machine.

        """
        # This is being run so python exists
        return True

    @classmethod
    def is_library_installed(cls, lib, **kwargs):
        r"""Determine if a dependency is installed.

        Args:
            lib (str): Name of the library that should be checked.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            bool: True if the library is installed, False otherwise.

        """
        try:
            importlib.import_module(lib)
        except ImportError:
            return False
        return True

    @classmethod
    def is_comm_installed(cls, **kwargs):
        r"""Determine if a comm is installed for the associated programming
        language.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            bool: True if a comm is installed for this language.

        """
        # Call __func__ to avoid direct invoking of class which dosn't exist
        # in after_registration where this is called
        out = InterpretedModelDriver.is_comm_installed.__func__(cls, **kwargs)
        if not kwargs.get('skip_config'):
            return out
        if out and (kwargs.get('commtype', None) in ['rmq', 'rmq_async']):
            from yggdrasil.communication.RMQComm import check_rmq_server
            out = check_rmq_server()
        return out



from jsonschema.compat import iteritems
from yggdrasil.metaschema import normalizer as normalizer_mod
from yggdrasil.metaschema.datatypes import encode_type, compare_schema
from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty


class PropertiesMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'properties' property."""

    name = 'properties'
    _replaces_existing = True
    _validate = False

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'properties' container property."""
        if typedef is None:
            typedef = {}
        return {k: encode_type(v, typedef=typedef.get(k, None))
                for k, v in instance.items()}

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Comparison method for 'properties' container property."""
        for k in prop2.keys():
            if k not in prop1:
                yield "Missing property '%s'" % k
                continue
            for e in compare_schema(prop1[k], prop2[k], root1=root1, root2=root2):
                yield e

    @classmethod
    def normalize(cls, validator, value, instance, schema):
        r"""Normalization method for 'properties' container property."""
        if not isinstance(instance, dict):
            return instance
        for property, subschema in iteritems(value):
            if property not in instance:
                instance[property] = normalizer_mod.UndefinedProperty()
        return instance

    @classmethod
    def post_validate(cls, validator, value, instance, schema):
        r"""Actions performed after validation if normalizing."""
        if not isinstance(instance, dict):
            return
        norm_keys = list(validator._normalized.keys())
        for k in norm_keys:
            if isinstance(validator._normalized[k], normalizer_mod.UndefinedProperty):
                del validator._normalized[k]



from yggdrasil.drivers.ServerDriver import ServerDriver


class RMQServerDriver(ServerDriver):
    r"""Class for handling an RMQ server.

    Args:
        *args: Arguments are passed to parent class.
        **kwargs: Keyword arguments are passed to parent class.

    """

    _connection_type = 'rmq_server'
    _schema_subtype_description = ('Connection between a RabbitMQ server request '
                                   'comm and a model acting as a server.')
    
    def __init__(self, *args, **kwargs):
        kwargs['comm'] = 'RMQComm'
        super(RMQServerDriver, self).__init__(*args, **kwargs)



import importlib
from yggdrasil import backwards
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType


class ClassMetaschemaType(MetaschemaType):
    r"""Type for evaluating classes."""

    name = 'class'
    description = 'Type for Python classes.'
    python_types = backwards.class_types
    encoded_type = 'string'
    cross_language_support = False

    @classmethod
    def normalize(cls, obj):
        r"""Normalize an object, if possible, to conform to this type.

        Args:
            obj (object): Object to normalize.

        Returns:
            object: Normalized object.

        """
        if isinstance(obj, backwards.string_types):
            try:
                obj = cls.decode_data(backwards.as_str(obj), {'type': cls.name})
            except (ValueError, AttributeError):
                pass
        return obj

    @classmethod
    def encode_data(cls, obj, typedef):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        mod = obj.__module__
        fun = obj.__name__
        out = '%s:%s' % (mod, fun)
        return out

    @classmethod
    def decode_data(cls, obj, typedef):
        r"""Decode an object.

        Args:
            obj (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        if not isinstance(obj, backwards.string_types):
            return obj
        pkg_mod = obj.split(':')
        if len(pkg_mod) != 2:
            raise ValueError("Could not parse %s string: %s"
                             % (cls.name, obj))
        mod, fun = pkg_mod[:]
        modobj = importlib.import_module(mod)
        if not hasattr(modobj, fun):
            raise AttributeError("Module %s has no %s %s"
                                 % (modobj, cls.name, fun))
        return getattr(modobj, fun)



import os
import unittest
from yggdrasil import tools
from yggdrasil.examples.tests import TestExample


_default_comm = tools.get_default_comm()


class ExampleTimedPipeTestBase(TestExample):
    r"""Base class for testing TimedPipe example with various comm types."""

    example_name = 'timed_pipe'
    env = {'PIPE_MSG_COUNT': '10',
           'PIPE_MSG_SIZE': '1024'}

    def __init__(self, *args, **kwargs):
        super(ExampleTimedPipeTestBase, self).__init__(*args, **kwargs)
        self._new_default_comm = getattr(self.__class__, '__new_default_comm',
                                         _default_comm)
        # if self._new_default_comm in ['ipc', 'IPCComm']:
        #     self.debug_flag = True
        
    def run_example(self):
        r"""This runs an example in the correct language."""
        if self._new_default_comm in ['ipc', 'IPCComm']:
            from yggdrasil.communication.IPCComm import ipcrm_queues, ipc_queues
            qlist = ipc_queues()
            if qlist:  # pragma: debug
                print('Existing queues:', qlist)
                ipcrm_queues()
        super(ExampleTimedPipeTestBase, self).run_example()

    @property
    def description_prefix(self):
        r"""Prefix message with test name."""
        out = super(ExampleTimedPipeTestBase, self).description_prefix
        out += '(%s)' % self._new_default_comm
        return out
    
    @property
    def results(self):
        r"""Result that should be found in output files."""
        siz = int(self.env['PIPE_MSG_COUNT']) * int(self.env['PIPE_MSG_SIZE'])
        res = '0' * siz
        return [res]

    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.tempdir, 'output_timed_pipe.txt')]


# Dynamically add test classes for comm types
for c in tools.get_installed_comm():
    if c == _default_comm:
        continue
    new_cls = unittest.skipIf(not tools.is_comm_installed(c),
                              "%s library not installed." % c)(
                                  type('TestExampleTimedPipe_%s' % c,
                                       (ExampleTimedPipeTestBase, ),
                                       {'__new_default_comm': c}))
    globals()[new_cls.__name__] = new_cls
    del new_cls



import unittest
import copy
from yggdrasil.tests import assert_raises
from yggdrasil.config import ygg_cfg
from yggdrasil.communication import new_comm
from yggdrasil.communication.RMQComm import RMQComm, check_rmq_server
from yggdrasil.communication.tests import test_AsyncComm


_rmq_installed = RMQComm.is_installed(language='python')


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQComm(test_AsyncComm.TestAsyncComm):
    r"""Test for RMQComm communication class."""

    comm = 'RMQComm'
    timeout = 10.0
    attr_list = (copy.deepcopy(test_AsyncComm.TestAsyncComm.attr_list)
                 + ['connection', 'channel'])

    
@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
def test_running():
    r"""Test checking for server w/ URL."""
    # url = "amqp://guest:guest@localhost:5672/%2F"
    vhost = ygg_cfg.get('rmq', 'vhost', '')
    if not vhost:
        vhost = '%2F'
    url = "amqp://%s:%s@%s:%s/%s" % (
        ygg_cfg.get('rmq', 'user', 'guest'),
        ygg_cfg.get('rmq', 'password', 'guest'),
        ygg_cfg.get('rmq', 'host', 'localhost'),
        ygg_cfg.get('rmq', 'port', '5672'),
        vhost)
    assert(check_rmq_server(url=url))
    

@unittest.skipIf(_rmq_installed, "RMQ Server running")
def test_not_running():
    r"""Test raise of an error if a RMQ server is not running."""
    comm_kwargs = dict(comm='RMQComm', direction='send', reverse_names=True)
    assert_raises(RuntimeError, new_comm, 'test', **comm_kwargs)



import os
import numpy as np
import unittest
from yggdrasil.communication import get_comm
from yggdrasil.interface import YggInterface
from yggdrasil.tools import YGG_MSG_EOF, get_YGG_MSG_MAX, YGG_MSG_BUF
from yggdrasil.components import import_component
from yggdrasil.drivers import (InputDriver, OutputDriver)
from yggdrasil.drivers.MatlabModelDriver import MatlabModelDriver
from yggdrasil.tests import YggTestClassInfo, assert_equal, assert_raises


YGG_MSG_MAX = get_YGG_MSG_MAX()


def test_maxMsgSize():
    r"""Test max message size."""
    assert_equal(YggInterface.maxMsgSize(), YGG_MSG_MAX)


def test_eof_msg():
    r"""Test eof message signal."""
    assert_equal(YggInterface.eof_msg(), YGG_MSG_EOF)


def test_bufMsgSize():
    r"""Test buf message size."""
    assert_equal(YggInterface.bufMsgSize(), YGG_MSG_BUF)


def test_init():
    r"""Test error on init."""
    assert_raises(Exception, YggInterface.YggInput, 'error')
    assert_raises(Exception, YggInterface.YggOutput, 'error')
    

@unittest.skipIf(not MatlabModelDriver.is_installed(), "Matlab not installed.")
def test_YggMatlab_class():  # pragma: matlab
    r"""Test Matlab interface for classes."""
    name = 'test'
    # Input
    drv = InputDriver.InputDriver(name, 'link')
    drv.start()
    os.environ.update(drv.env)
    YggInterface.YggMatlab('YggInput', (name, '%f\\n%d'))
    drv.terminate()
    # Output
    drv = OutputDriver.OutputDriver(name, 'link')
    drv.start()
    os.environ.update(drv.env)
    YggInterface.YggMatlab('YggOutput', (name, '%f\\n%d'))
    drv.terminate()


@unittest.skipIf(not MatlabModelDriver.is_installed(), "Matlab not installed.")
def test_YggMatlab_variables():  # pragma: matlab
    r"""Test Matlab interface for variables."""
    assert_equal(YggInterface.YggMatlab('YGG_MSG_MAX'), YGG_MSG_MAX)
    assert_equal(YggInterface.YggMatlab('YGG_MSG_EOF'), YGG_MSG_EOF)


class TestBase(YggTestClassInfo):
    r"""Test class for interface classes."""
    def __init__(self, *args, **kwargs):
        super(TestBase, self).__init__(*args, **kwargs)
        self._mod = 'yggdrasil.interface.YggInterface'
        self.name = 'test' + self.uuid
        self.matlab = False
        self.idriver = None
        self.odriver = None
        self.test_comm = None
        self.is_file = False
        self.filecomm = None
        self.filename = os.path.join(os.getcwd(), 'temp_ascii.txt')
        self.testing_option_kws = {}
        self.direction = None
        self.test_comm_kwargs = {}
        # self._driver_kwargs = {}
        self._inst_args = [self.name]
        self.fmt_str = b'%5s\t%d\t%f\n'
        self.fmt_str_matlab = b'%5s\\t%d\\t%f\\n'

    @property
    def odriver_class(self):
        r"""class: Output driver class."""
        if self.direction is None:
            return None  # pragma: no cover
        elif (self.direction == 'output') and self.is_file:
            return import_component('connection', 'file_output')
        elif (self.direction == 'input') and self.is_file:
            return None
        return import_component('connection', 'output')

    @property
    def idriver_class(self):
        r"""class: Input driver class."""
        if self.direction is None:
            return None  # pragma: no cover
        elif (self.direction == 'output') and self.is_file:
            return None
        elif (self.direction == 'input') and self.is_file:
            return import_component('connection', 'file_input')
        return import_component('connection', 'input')

    @property
    def odriver_args(self):
        r"""list: Output driver arguments."""
        if (self.direction == 'output') and self.is_file:
            filecomm_kwargs = self.testing_options['kwargs']
            filecomm_kwargs['comm'] = self.filecomm
            return ([self.name, self.filename],
                    {'ocomm_kws': filecomm_kwargs})
        elif (self.direction == 'input') and self.is_file:
            return None, None  # pragma: no cover
        elif (self.direction == 'output'):
            return [self.name, self.name + '_link'], {}
        elif (self.direction == 'input'):
            return [self.name + '_odriver', self.name + '_link'], {}
        raise Exception('Direction was not set. (%s)', self.direction)  # pragma: debug

    @property
    def idriver_args(self):
        r"""list: Input driver arguments."""
        if (self.direction == 'output') and self.is_file:
            return None, None  # pragma: no cover
        elif (self.direction == 'input') and self.is_file:
            filecomm_kwargs = self.testing_options['kwargs']
            filecomm_kwargs['comm'] = self.filecomm
            return ([self.name, self.filename],
                    {'icomm_kws': filecomm_kwargs})
        elif (self.direction == 'output'):
            return [self.name + '_idriver', self.name + '_link'], {}
        elif (self.direction == 'input'):
            return [self.name, self.name + '_link'], {}
        raise Exception('Direction was not set. (%s)', self.direction)  # pragma: debug

    @property
    def inst_kwargs(self):
        r"""dict: Arguments for the interface instance."""
        out = super(TestBase, self).inst_kwargs
        out['matlab'] = self.matlab
        return out
        
    def get_options(self):
        r"""Get testing options."""
        out = {}
        if self.is_file:
            assert(self.filecomm is not None)
            out = import_component('file', self.filecomm).get_testing_options(
                **self.testing_option_kws)
        else:
            out = import_component('comm', 'default').get_testing_options(
                **self.testing_option_kws)
        return out

    @property
    def messages(self):
        r"""list: Messages that should be sent/received."""
        if getattr(self, '_messages', None) is not None:
            return self._messages
        return self.testing_options['send']

    def setup(self):
        r"""Start driver and instance."""
        if self.direction is None:  # pragma: debug
            return
        nprev_comm = self.comm_count
        nprev_thread = self.thread_count
        nprev_fd = self.fd_count
        idriver_class = self.idriver_class
        odriver_class = self.odriver_class
        # File
        if self.is_file and (self.direction == 'input'):
            with open(self.filename, 'wb') as fd:
                fd.write(self.testing_options['contents'])
        # Drivers
        comm_env = None
        if idriver_class is not None:
            args, kwargs = self.idriver_args
            self.idriver = idriver_class(*args, **kwargs)
            self.idriver.start()
            comm_env = self.idriver.comm_env
        if odriver_class is not None:
            args, kwargs = self.odriver_args
            if comm_env is not None:
                kwargs['comm_env'] = comm_env
            self.odriver = odriver_class(*args, **kwargs)
            self.odriver.start()
        # Test comm
        if self.direction == 'input':
            os.environ.update(self.idriver.env)
            if self.odriver is not None:
                kws = self.odriver.icomm.opp_comm_kwargs()
                kws.update(self.test_comm_kwargs)
                self.test_comm = get_comm('in', **kws)
        elif self.direction == 'output':
            os.environ.update(self.odriver.env)
            if self.idriver is not None:
                kws = self.idriver.ocomm.opp_comm_kwargs()
                kws.update(self.test_comm_kwargs)
                self.test_comm = get_comm('out', **kws)
        # Test class
        super(TestBase, self).setup(nprev_comm=nprev_comm,
                                    nprev_thread=nprev_thread,
                                    nprev_fd=nprev_fd)

    def teardown(self):
        r"""Stop the driver."""
        if self.odriver is not None:
            self.odriver.terminate()
            self.odriver.cleanup()
        if self.idriver is not None:
            self.idriver.terminate()
            self.idriver.cleanup()
        if self.test_comm is not None:
            self.test_comm.close()
        if self.is_file and os.path.isfile(self.filename):
            os.remove(self.filename)
        if self.direction is None:  # pragma: debug
            return
        super(TestBase, self).teardown()
        self.cleanup_comms()

    def remove_instance(self, inst):
        r"""Remove an instance."""
        inst.is_interface = False
        inst.close()
        assert(inst.is_closed)
        super(TestBase, self).remove_instance(inst)
            
    
class TestYggInput(TestBase):
    r"""Test basic input to python."""
    def __init__(self, *args, **kwargs):
        super(TestYggInput, self).__init__(*args, **kwargs)
        self._cls = 'YggInput'
        self.direction = 'input'
        if self.__class__ == TestYggInput:
            self.testing_option_kws = {'table_example': True}
            self._inst_kwargs = {'format_str': self.fmt_str}

    def test_msg(self):
        r"""Test sending/receiving message."""
        if self.is_file:
            for msg in self.testing_options['recv']:
                msg_flag, msg_recv = self.instance.recv(self.timeout)
                assert(msg_flag)
                self.assert_equal(msg_recv, msg)
            msg_flag, msg_recv = self.instance.recv(self.timeout)
            assert(not msg_flag)
        else:
            for msg in self.messages:
                msg_flag = self.test_comm.send(msg)
                assert(msg_flag)
                msg_flag, msg_recv = self.instance.recv(self.timeout)
                assert(msg_flag)
                self.assert_equal(msg_recv, msg)
            

class TestYggInputMatlab(TestYggInput):
    r"""Test basic input to python as passed from matlab."""
    def __init__(self, *args, **kwargs):
        super(TestYggInputMatlab, self).__init__(*args, **kwargs)
        self.matlab = True
        self.testing_option_kws = {'table_example': True}
        self._inst_kwargs = {'format_str': self.fmt_str_matlab}


class TestYggOutput(TestBase):
    r"""Test basic output to python."""
    def __init__(self, *args, **kwargs):
        super(TestYggOutput, self).__init__(*args, **kwargs)
        self._cls = 'YggOutput'
        self.direction = 'output'
        if self.__class__ == TestYggOutput:
            self.testing_option_kws = {'table_example': True}
            self._inst_kwargs = {'format_str': self.fmt_str}

    def test_msg(self):
        r"""Test sending/receiving message."""
        if self.is_file:
            for msg in self.testing_options['send']:
                msg_flag = self.instance.send(msg)
                assert(msg_flag)
            self.instance.send_eof()
            # Read temp file
            Tout = self.instance.start_timeout()
            while self.odriver.ocomm.is_open and not Tout.is_out:
                self.instance.sleep()
            self.instance.stop_timeout()
            assert(os.path.isfile(self.filename))
            if self.testing_options.get('exact_contents', True):
                with open(self.filename, 'rb') as fd:
                    res = fd.read()
                    self.assert_equal(res, self.testing_options['contents'])
        else:
            for msg in self.messages:
                msg_flag = self.instance.send(msg)
                assert(msg_flag)
                msg_flag, msg_recv = self.test_comm.recv(self.timeout)
                assert(msg_flag)
                self.assert_equal(msg_recv, msg)
        

class TestYggOutputMatlab(TestYggOutput):
    r"""Test basic output to python as passed from matlab."""
    def __init__(self, *args, **kwargs):
        super(TestYggOutputMatlab, self).__init__(*args, **kwargs)
        self.matlab = True
        self.testing_option_kws = {'table_example': True}
        self._inst_kwargs = {'format_str': self.fmt_str_matlab}


class TestYggRpcClient(TestYggOutput):
    r"""Test client-side RPC communication with Python."""
    def __init__(self, *args, **kwargs):
        super(TestYggRpcClient, self).__init__(*args, **kwargs)
        self._cls = 'YggRpcClient'
        self._inst_args = [self.name, self.fmt_str, self.fmt_str]
        self.test_comm_kwargs = {'comm': 'ServerComm',
                                 'response_kwargs': {'format_str': self.fmt_str}}
        self._messages = [(b'one', np.int32(1), 1.0)]
        
    @property
    def odriver_class(self):
        r"""class: Output driver class."""
        return import_component('connection', 'client')

    @property
    def idriver_class(self):
        r"""class: Input driver class."""
        return import_component('connection', 'server')
    
    def test_msg(self):
        r"""Test sending/receiving message."""
        super(TestYggRpcClient, self).test_msg()
        for msg in self.messages:
            msg_flag = self.test_comm.send(msg)
            assert(msg_flag)
            msg_flag, msg_recv = self.instance.recv(self.timeout)
            assert(msg_flag)
            self.assert_equal(msg_recv, msg)
        
        
class TestYggRpcClientMatlab(TestYggRpcClient):
    r"""Test client-side RPC communication with Python as passed through Matlab."""
    def __init__(self, *args, **kwargs):
        super(TestYggRpcClientMatlab, self).__init__(*args, **kwargs)
        self.matlab = True
        self._inst_args = [self.name, self.fmt_str_matlab, self.fmt_str_matlab]


class TestYggRpcServer(TestYggInput):
    r"""Test server-side RPC communication with Python."""
    def __init__(self, *args, **kwargs):
        super(TestYggRpcServer, self).__init__(*args, **kwargs)
        self._cls = 'YggRpcServer'
        self._inst_args = [self.name, self.fmt_str, self.fmt_str]
        self.test_comm_kwargs = {'comm': 'ClientComm',
                                 'response_kwargs': {'format_str': self.fmt_str}}
        self._messages = [(b'one', np.int32(1), 1.0)]
        
    @property
    def odriver_class(self):
        r"""class: Output driver class."""
        return import_component('connection', 'client')

    @property
    def idriver_class(self):
        r"""class: Input driver class."""
        return import_component('connection', 'server')
    
    def test_msg(self):
        r"""Test sending/receiving message."""
        super(TestYggRpcServer, self).test_msg()
        for msg in self.messages:
            msg_flag = self.instance.send(msg)
            assert(msg_flag)
            msg_flag, msg_recv = self.test_comm.recv(self.timeout)
            assert(msg_flag)
            self.assert_equal(msg_recv, msg)
        
        
class TestYggRpcServerMatlab(TestYggRpcServer):
    r"""Test server-side RPC communication with Python as passed through Matlab."""
    def __init__(self, *args, **kwargs):
        super(TestYggRpcServerMatlab, self).__init__(*args, **kwargs)
        self.matlab = True
        self._inst_args = [self.name, self.fmt_str_matlab, self.fmt_str_matlab]


# AsciiFile
class TestYggAsciiFileInput(TestYggInput):
    r"""Test input from an unformatted text file."""
    def __init__(self, *args, **kwargs):
        super(TestYggAsciiFileInput, self).__init__(*args, **kwargs)
        self._cls = 'YggAsciiFileInput'
        self.is_file = True
        self.filecomm = 'AsciiFileComm'


class TestYggAsciiFileOutput(TestYggOutput):
    r"""Test output to an unformatted text file."""
    def __init__(self, *args, **kwargs):
        super(TestYggAsciiFileOutput, self).__init__(*args, **kwargs)
        self._cls = 'YggAsciiFileOutput'
        self.is_file = True
        self.filecomm = 'AsciiFileComm'


# AsciiTable
class TestYggAsciiTableInput(TestYggAsciiFileInput):
    r"""Test input from an ascii table."""
    def __init__(self, *args, **kwargs):
        super(TestYggAsciiTableInput, self).__init__(*args, **kwargs)
        self._cls = 'YggAsciiTableInput'
        self.is_file = True
        self.filecomm = 'AsciiTableComm'

        
class TestYggAsciiTableOutput(TestYggAsciiFileOutput):
    r"""Test output from an ascii table."""
    def __init__(self, *args, **kwargs):
        super(TestYggAsciiTableOutput, self).__init__(*args, **kwargs)
        self._cls = 'YggAsciiTableOutput'
        self.is_file = True
        self.filecomm = 'AsciiTableComm'
        self._inst_args = [self.name, self.fmt_str]
        self._inst_kwargs = {}


class TestYggAsciiTableOutputMatlab(TestYggAsciiTableOutput):
    r"""Test output from an ascii table as passed through Matlab."""
    def __init__(self, *args, **kwargs):
        super(TestYggAsciiTableOutputMatlab, self).__init__(*args, **kwargs)
        self.matlab = True
        self._inst_args = [self.name, self.fmt_str_matlab]
        

# AsciiTable Array
class TestYggAsciiArrayInput(TestYggAsciiTableInput):
    r"""Test input from an ASCII table."""
    def __init__(self, *args, **kwargs):
        super(TestYggAsciiArrayInput, self).__init__(*args, **kwargs)
        self._cls = 'YggAsciiArrayInput'
        self.is_file = True
        self.filecomm = 'AsciiTableComm'
        self.testing_option_kws = {'array_columns': True}


class TestYggAsciiArrayOutput(TestYggAsciiTableOutput):
    r"""Test input from an ASCII table."""
    def __init__(self, *args, **kwargs):
        super(TestYggAsciiArrayOutput, self).__init__(*args, **kwargs)
        self._cls = 'YggAsciiArrayOutput'
        self.is_file = True
        self.filecomm = 'AsciiTableComm'
        self.testing_option_kws = {'array_columns': True}
        

# Pickle
class TestYggPickleInput(TestYggInput):
    r"""Test input from a pickle file."""
    def __init__(self, *args, **kwargs):
        super(TestYggPickleInput, self).__init__(*args, **kwargs)
        self._cls = 'YggPickleInput'
        self.is_file = True
        self.filecomm = 'PickleFileComm'


class TestYggPickleOutput(TestYggOutput):
    r"""Test output from a pickle."""
    def __init__(self, *args, **kwargs):
        super(TestYggPickleOutput, self).__init__(*args, **kwargs)
        self._cls = 'YggPickleOutput'
        self.is_file = True
        self.filecomm = 'PickleFileComm'

        
# Pandas
class TestYggPandasInput(TestYggInput):
    r"""Test input from a pandas file."""
    def __init__(self, *args, **kwargs):
        super(TestYggPandasInput, self).__init__(*args, **kwargs)
        self._cls = 'YggPandasInput'
        self.is_file = True
        self.filecomm = 'PandasFileComm'
        # self.testing_option_kws = {'as_frames': True}


class TestYggPandasOutput(TestYggOutput):
    r"""Test output from a pandas."""
    def __init__(self, *args, **kwargs):
        super(TestYggPandasOutput, self).__init__(*args, **kwargs)
        self._cls = 'YggPandasOutput'
        self.is_file = True
        self.filecomm = 'PandasFileComm'
        # self.testing_option_kws = {'as_frames': True}


# Ply
class TestYggPlyInput(TestYggInput):
    r"""Test input from a ply file."""
    def __init__(self, *args, **kwargs):
        super(TestYggPlyInput, self).__init__(*args, **kwargs)
        self._cls = 'YggPlyInput'
        self.is_file = True
        self.filecomm = 'PlyFileComm'


class TestYggPlyOutput(TestYggOutput):
    r"""Test output from a ply."""
    def __init__(self, *args, **kwargs):
        super(TestYggPlyOutput, self).__init__(*args, **kwargs)
        self._cls = 'YggPlyOutput'
        self.is_file = True
        self.filecomm = 'PlyFileComm'


# Obj
class TestYggObjInput(TestYggInput):
    r"""Test input from a obj file."""
    def __init__(self, *args, **kwargs):
        super(TestYggObjInput, self).__init__(*args, **kwargs)
        self._cls = 'YggObjInput'
        self.is_file = True
        self.filecomm = 'ObjFileComm'


class TestYggObjOutput(TestYggOutput):
    r"""Test output from a obj."""
    def __init__(self, *args, **kwargs):
        super(TestYggObjOutput, self).__init__(*args, **kwargs)
        self._cls = 'YggObjOutput'
        self.is_file = True
        self.filecomm = 'ObjFileComm'



import warnings
from yggdrasil.interface import YggInterface
warnings.warn(("YggInterface replaces CisInterface in yggdrasil. "
               "Replace ygg with cis in interface calls."),
              DeprecationWarning, stacklevel=2)


direct_replace = ['maxMsgSize', 'bufMsgSize', 'eof_msg']
upper_replace = ['YGG_MSG_MAX', 'YGG_MSG_EOF', 'YGG_MSG_BUF']
title_replace = ['YggMatlab', 'YggInput', 'YggOutput',
                 'YggRpcServer', 'YggRpcClient',
                 'YggAsciiFileInput', 'YggAsciiFileOutput',
                 'YggAsciiTableInput', 'YggAsciiTableOutput',
                 'YggAsciiArrayInput', 'YggAsciiArrayOutput',
                 'YggPickleInput', 'YggPickleOutput',
                 'YggPandasInput', 'YggPandasOutput',
                 'YggPlyInput', 'YggPlyOutput',
                 'YggObjInput', 'YggObjOutput']


for x in direct_replace:
    globals()[x] = getattr(YggInterface, x)
for x in upper_replace:
    globals()[x.replace('YGG', 'CIS')] = getattr(YggInterface, x)
for x in title_replace:
    globals()[x.replace('Ygg', 'Cis')] = getattr(YggInterface, x)



import numpy as np
from yggdrasil import backwards, units, serialize
from yggdrasil.serialize.SerializeBase import SerializeBase


class DefaultSerialize(SerializeBase):
    r"""Default class for serializing/deserializing a python object into/from
    a bytes message.

    Args:
        **kwargs: Additional keyword args are passed to the parent class.

    """

    _seritype = 'default'
    _schema_subtype_description = ('Default serializer that uses |yggdrasil|\'s '
                                   'extended JSON serialization based on a '
                                   'provided type definition (See discussion '
                                   ':ref:`here <serialization_rst>`).')
    func_serialize = None
    func_deserialize = None
    
    @classmethod
    def object2dict(cls, obj, as_array=False, field_names=None, **kwargs):
        r"""Convert a message object into a dictionary.

        Args:
            obj (object): Object that would be serialized by this class and
                should be returned in a dictionary form.
            as_array (bool, optional): If True, the objects in the list
                are complete columns in a table and as_format is set to True.
                Defaults to False.
            field_names (list, optional): The field names associated with a
                table-like data type. Defaults to None. This keyword must be
                provided if as_array is True.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            dict: Dictionary version of the provided object.

        """
        if field_names is None:
            assert(not as_array)
            return super(DefaultSerialize, cls).object2dict(obj, **kwargs)
        else:
            out = serialize.list2dict(obj, names=field_names)
        return out

    @classmethod
    def object2array(cls, obj, as_array=False, field_names=None, **kwargs):
        r"""Convert a message object into an array.

        Args:
            obj (object): Object that would be serialized by this class and
                should be returned in an array form.
            as_array (bool, optional): If True, the objects in the list
                are complete columns in a table and as_format is set to True.
                Defaults to False.
            field_names (list, optional): The field names associated with a
                table-like data type. Defaults to None. This keyword must be
                provided if as_array is True.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            np.array: Array version of the provided object.

        """
        if as_array:
            assert(field_names is not None)
            out = serialize.list2numpy(obj, names=field_names)
        else:
            out = super(DefaultSerialize, cls).object2array(obj, **kwargs)
        return out

    @classmethod
    def concatenate(cls, objects, as_array=False, **kwargs):
        r"""Concatenate objects to get object that would be recieved if
        the concatenated serialization were deserialized.

        Args:
            objects (list): Objects to be concatenated.
            as_array (bool, optional): If True, the objects in the list
                are complete columns in a table and as_format is set to True.
                Defaults to False.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Set of objects that results from concatenating those provided.

        """
        if len(objects) == 0:
            return []
        if as_array:
            units_list = [units.get_units(ix) for ix in objects[0]]
            out = [[units.add_units(np.hstack([x[i] for x in objects]), u)
                    for i, u in enumerate(units_list)]]
        elif isinstance(objects[0], backwards.bytes_type):
            out = [b''.join(objects)]
        else:
            return super(DefaultSerialize, cls).concatenate(objects, **kwargs)
        return out

    @classmethod
    def get_testing_options(cls, **kwargs):
        r"""Method to return a dictionary of testing options for this class."""
        out = super(DefaultSerialize, cls).get_testing_options(**kwargs)
        if cls._seritype == 'default':
            out['concatenate'] = [([], []),
                                  ([b'a', b'b'], [b'ab'])]
        return out
        
    def update_serializer(self, *args, **kwargs):
        r"""Update serializer with provided information.

        Args:
            *args: All arguments are passed to the parent class's method.
            **kwargs: All keyword arguments are passed to the parent class's
                method.

        """
        out = super(DefaultSerialize, self).update_serializer(*args, **kwargs)
        if (self.func_serialize is None) or (self.func_deserialize is None):
            self.encoded_datatype = self.datatype
        return out



import numpy as np
import logging
from yggdrasil import backwards
if backwards.PY2:  # pragma: Python 2
    import pint
    unyt = None
    _use_unyt = False
    _ureg_unyt = None
    _ureg_pint = pint.UnitRegistry()
    # _ureg_pint.define('micro_mole = 1e-6 * mole = uMol = umol')
    # Don't print warning about day redefinition
    logging.getLogger("pint").propagate = False
    _ureg_pint.define('day = 24 * hour = d')
    logging.getLogger("pint").propagate = True
else:  # pragma: Python 3
    import unyt
    pint = None
    _use_unyt = True
    _ureg_unyt = unyt.UnitRegistry()
    _ureg_pint = None
if _use_unyt:
    _unit_quantity = unyt.array.unyt_quantity
    _unit_array = unyt.array.unyt_array
else:
    _unit_quantity = _ureg_pint.Quantity
    _unit_array = _ureg_pint.Quantity


def has_units(obj):
    r"""Determine if a Python object has associated units.

    Args:
        obj (object): Object to be tested for units.

    Returns:
        bool: True if the object has units, False otherwise.

    """
    return hasattr(obj, 'units')


def get_units(obj):
    r"""Get the string representation of the units.

    Args:
        obj (object): Object to get units for.

    Returns:
        str: Units, empty if input object has none.
        
    """
    if has_units(obj):
        out = str(obj.units)
    else:
        out = ''
    return out


def get_data(obj):
    r"""Get the array/scalar assocaited with the object.

    Args:
        obj (object): Object to get data for.

    Returns:
        np.ndarray: Numpy array representation of the underlying data.

    """
    if has_units(obj):
        if _use_unyt:
            out = obj.to_ndarray()
            if out.ndim == 0:
                out = out.reshape((1, ))[0]
        else:
            out = obj.magnitude
    else:
        out = obj
    return out


def add_units(arr, unit_str, dtype=None):
    r"""Add units to an array or scalar.

    Args:
        arr (np.ndarray, float, int): Scalar or array of data to add units to.
        unit_str (str): Unit string.
        dtype (np.dtype, optional): Numpy data type that should be maintained for
            array/qunatity with units. If not provided, this is determined from the
            array.

    Returns:
        unyt.unyt_array: Array with units.

    """
    if isinstance(unit_str, backwards.bytes_type):
        unit_str = backwards.as_str(unit_str)
    if is_null_unit(unit_str):
        return arr
    if dtype is None:
        if isinstance(arr, np.ndarray):
            dtype = arr.dtype
        else:
            dtype = np.array([arr]).dtype
    if _use_unyt:
        if isinstance(arr, np.ndarray) and (arr.ndim > 0):
            out = unyt.unyt_array(arr, unit_str, dtype=dtype)
        else:
            out = unyt.unyt_quantity(arr, unit_str, dtype=dtype)
    else:
        out = _ureg_pint.Quantity(arr, unit_str)
    return out


def are_compatible(units1, units2):
    r"""Check if two units are compatible.

    Args:
        units1 (str): First units string.
        units2 (str): Second units string.

    Returns:
        bool: True if the units are compatible, False otherwise.

    """
    # Empty units always compatible
    if is_null_unit(units1) or is_null_unit(units2):
        return True
    if (not is_unit(units1)) or (not is_unit(units2)):
        return False
    x = add_units(1, units1)
    try:
        convert_to(x, units2)
    except ValueError:
        return False
    return True


def is_null_unit(ustr):
    r"""Determines if a string is a null unit.

    Args:
        ustr (str): String to test.

    Returns:
        bool: True if the string is '' or 'n/a', False otherwise.

    """
    if (len(ustr) == 0) or (ustr == 'n/a'):
        return True
    return False


def as_unit(ustr):
    r"""Get unit object for the string.

    Args:

    Returns:

    Raises:
        ValueError: If the string is not a recognized unit.

    """
    if _use_unyt:
        try:
            out = unyt.Unit(ustr)
        except unyt.exceptions.UnitParseError as e:
            raise ValueError(str(e))
    else:
        try:
            out = _ureg_pint(ustr)
        except pint.errors.UndefinedUnitError as e:
            raise ValueError(str(e))
    return out


def is_unit(ustr):
    r"""Determine if a string is a valid unit.

    Args:
        ustr (str): String representation to test.

    Returns:
        bool: True if the string is a valid unit. False otherwise.

    """
    ustr = backwards.as_str(ustr)
    if is_null_unit(ustr):
        return True
    try:
        as_unit(ustr)
    except ValueError:
        return False
    return True


def convert_to(arr, new_units):
    r"""Convert qunatity with units to new units. Objects without units
    will be returned with the new units.

    Args:
        arr (np.ndarray, float, int, unyt.unyt_array): Quantity with or
            without units.
        new_units (str): New units that should be applied.

    Returns:
        unyt.unyt_array: Array with new units.

    """
    if is_null_unit(new_units):
        return arr
    if not has_units(arr):
        return add_units(arr, new_units)
    if _use_unyt:
        try:
            out = arr.to(new_units)
        except unyt.exceptions.UnitConversionError as e:
            raise ValueError(str(e))
    else:
        out = arr.to(new_units)
    return out



import os
from yggdrasil.tests import extra_example
from yggdrasil.examples.tests import TestExample


@extra_example
class TestExampleGS4b(TestExample):
    r"""Test the Getting Started Lesson 4b example."""

    example_name = 'gs_lesson4b'

    @property
    def input_files(self):
        r"""Input file."""
        return [os.path.join(self.yamldir, 'Input', 'input.txt')]

    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.yamldir, 'output.txt')]



import threading
from yggdrasil import backwards, tools
from yggdrasil.communication import RMQComm
if RMQComm._rmq_installed:
    import pika
    _pika_version_maj = int(float(pika.__version__.split('.')[0]))
    if _pika_version_maj >= 1:  # pragma: debug
        raise ImportError("pika version 1.0 not yet supported.")
else:
    pika = False
    _pika_version_maj = 0


class RMQAsyncComm(RMQComm.RMQComm):
    r"""Class for handling asynchronous RabbitMQ communications. It is not
    recommended to use this class as it can leave hanging threads if not
    closed propertly. The normal RMQComm will cover most use cases.

    Args:
        name (str): The environment variable where the comm address is stored.
        dont_open (bool, optional): If True, the connection will not be opened.
            Defaults to False.
        **kwargs: Additional keyword arguments are passed to CommBase.

    Attributes:
        times_connected (int): Number of times that this connections has been
            established.
        rmq_thread (tools.YggThread): Thread used to run IO loop.

    """
    
    _commtype = 'rmq_async'
    _schema_subtype_description = ('Asynchronous RabbitMQ connection.')
    
    def _init_before_open(self, **kwargs):
        r"""Initialize null variables and RMQ async thread."""
        self.times_connected = 0
        self.rmq_thread_count = 0
        self.rmq_thread = self.new_run_thread()
        self._opening = False
        self._closing = False
        self._reconnecting = False
        self._close_called = False
        self._buffered_messages = []
        self._qres = None
        self._qres_lock = threading.RLock()
        self._qres_event = threading.Event()
        self._qres_event.set()
        super(RMQAsyncComm, self)._init_before_open(**kwargs)

    @property
    def rmq_lock(self):
        r"""Lock associated with RMQ ioloop thread."""
        return self.rmq_thread.lock

    def new_run_thread(self, name=None):
        r"""Get a new thread for running."""
        if name is None:
            name = self.name
        self.rmq_thread_count += 1
        return tools.YggThread(name=name + '.RMQThread%d' % self.rmq_thread_count,
                               target=self.run_thread)

    def run_thread(self):
        r"""Connect to the connection and begin the IO loop."""
        self.debug('')
        self.connect()
        self.connection.ioloop.start()
        self.debug("returning")

    def start_run_thread(self):
        r"""Start the run thread and wait for it to finish."""
        with self.rmq_lock:
            if self.rmq_thread.was_started:
                return
            self._opening = True
            self.rmq_thread.start()
        # Wait for connection to be established
        T = self.start_timeout()
        # interval = 1  # timeout / 5
        while (not T.is_out) and (not self.channel_stable) and self.rmq_thread.is_alive():
            self.sleep()  # 0.5 # interval)
        self.stop_timeout()
        # Check that connection was established
        if not self.rmq_thread.is_alive():  # pragma: debug
            self._opening = False
            self.force_close()
            raise Exception("Connection ioloop could not be established.")
        if not self.channel_stable:  # pragma: debug
            self.force_close()
            raise RuntimeError("Connection never finished opening "
                               + "(%f/%f timeout)." % (T.elapsed, T.max_time))

    def bind(self):
        r"""Declare queue to get random new queue."""
        # Start ioloop in a new thread
        with self.rmq_lock:
            # Don't bind if already closing
            if self.is_open or self._close_called:  # pragma: debug
                return
            self._bound = True
        self.start_run_thread()
        # Register queue
        if not self.queue:  # pragma: debug
            self.error("Queue was not initialized.")
        self.register_comm(self.address, (self.connection, self.channel))
        super(RMQComm.RMQComm, self).bind()
    
    def _open_direct(self):
        r"""Open connection and bind/connect to queue as necessary."""
        super(RMQAsyncComm, self)._open_direct()
        T = self.start_timeout()
        while (not T.is_out) and self._opening:  # pragma: debug
            self.sleep()
        self.stop_timeout()

    def _close_direct(self, linger=False):
        r"""Close the connection.

        Args:
            linger (bool, optional): If True, drain messages before closing the
                comm. Defaults to False.

        """
        with self.rmq_lock:
            self._close_called = True
            if self._closing:  # pragma: debug
                return  # Don't close more than once
        # Wait for connection to finish opening to close it
        T = self.start_timeout(key=self.timeout_key + '_opening')
        while (not T.is_out) and self._opening:  # pragma: debug
            self.sleep()
        self.stop_timeout(key=self.timeout_key + '_opening')
        # Close by cancelling consumption
        if self.is_open or self._bound:
            with self.rmq_lock:
                self._closing = True
                self._is_open = False
                self._bound = False
                if self.channel is not None:
                    try:
                        if self.direction == 'recv':
                            self.channel.basic_cancel(callback=self.on_cancelok,
                                                      consumer_tag=self.consumer_tag)
                        else:
                            self.channel.close()
                    except (pika.exceptions.ChannelClosed,
                            pika.exceptions.ConnectionClosed):  # pragma: debug
                        self._closing = False
            if not self.is_client:
                self.unregister_comm(self.address)
        # Wait for connection to finish closing & then force if it dosn't
        T = self.start_timeout(key=self.timeout_key + '_closing')
        while (not T.is_out) and self._closing:
            self.sleep()
        self.stop_timeout(key=self.timeout_key + '_closing')
        if self._closing:  # pragma: debug
            self.force_close()
        if self.rmq_thread.is_alive():  # pragma: debug
            self.rmq_thread.join(self.timeout)
            if self.rmq_thread.is_alive():
                raise RuntimeError("Thread still running.")
        # Close workers
        # with self.rmq_lock:
        super(RMQAsyncComm, self)._close_direct(linger=linger)

    def _set_qres(self, res):
        r"""Callback for getting message count."""
        self._qres = res
        self._qres_event.set()

    def get_queue_result(self):
        r"""Get the fram from passive queue declare."""
        res = None
        if self.is_open_direct:
            with self._qres_lock:
                if self._qres_event.is_set():
                    self._qres_event.clear()
                    try:
                        self.channel.queue_declare(queue=self.queue,
                                                   callback=self._set_qres,
                                                   # , auto_delete=True,
                                                   passive=True)
                    except (pika.exceptions.ChannelClosed,
                            pika.exceptions.ConnectionClosed):  # pragma: debug
                        if not self._reconnecting:
                            self._close_direct()
                        else:
                            self._qres = None
                            self._qres_event.set()
            self._qres_event.wait()
            res = self._qres
        return res

    @property
    def n_msg_direct_recv(self):
        r"""int: Number of messages in the queue."""
        if self.is_open_direct:
            return len(self._buffered_messages)
        return 0
        
    # Access work comms with lock
    # def get_work_comm(self, *args, **kwargs):
    #     r"""Alias for parent class that wraps method in Lock."""
    #     with self.rmq_lock:
    #         return super(RMQAsyncComm, self).get_work_comm(*args, **kwargs)

    # def create_work_comm(self, *args, **kwargs):
    #     r"""Alias for parent class that wraps method in Lock."""
    #     with self.rmq_lock:
    #         return super(RMQAsyncComm, self).create_work_comm(*args, **kwargs)

    # def add_work_comm(self, *args, **kwargs):
    #     r"""Alias for parent class that wraps method in Lock."""
    #     with self.rmq_lock:
    #         return super(RMQAsyncComm, self).add_work_comm(*args, **kwargs)

    # def remove_work_comm(self, *args, **kwargs):
    #     r"""Alias for parent class that wraps method in Lock."""
    #     with self.rmq_lock:
    #         return super(RMQAsyncComm, self).remove_work_comm(*args, **kwargs)

    def _send_direct(self, msg, exchange=None, routing_key=None, **kwargs):
        r"""Send a message.

        Args:
            msg (str, bytes): Message to be sent.
            exchange (str, optional): Exchange that message should be sent
                to. Defaults to self.exchange.
            routing_key (str, optional): Key that exchange should use to route
                the message. Defaults to self.queue.
            **kwargs: Additional keyword arguments are passed to
                :method:`pika.BlockingChannel.basic_publish`.

        Returns:
            bool: Success or failure of send.

        """
        with self.rmq_lock:
            if self.is_closed:  # pragma: debug
                return False
            out = super(RMQAsyncComm, self)._send_direct(msg, exchange=exchange,
                                                         routing_key=routing_key,
                                                         **kwargs)
        # Basic publish returns None for asynchronous connection
        if out is None:
            out = True
        return out

    def _recv_direct(self):
        r"""Receive a message.

        Returns:
            tuple (bool, obj): Success or failure of receive and received
                message.

        """
        with self.rmq_lock:
            return (True, self._buffered_messages.pop(0))
        # if self.n_msg_recv != 0:
        #     with self.rmq_lock:
        #         out = (True, self._buffered_messages.pop(0))
        #     return out
        # if self.is_closed:  # pragma: debug
        #     self.debug("Connection closed.")
        #     return (False, None)
        # if self.n_msg_recv == 0:
        #     # self.debug(".recv(): No buffered messages.")
        #     out = (True, self.empty_bytes_msg)
        # else:
        #     with self.rmq_lock:
        #         out = (True, self._buffered_messages.pop(0))
        # return out

    def on_message(self, ch, method, props, body):
        r"""Buffer received messages."""
        if self.direction == 'send':  # pragma: debug
            raise Exception("Send comm received a message.")
        with self.rmq_lock:
            self._buffered_messages.append(backwards.as_bytes(body))
        ch.basic_ack(delivery_tag=method.delivery_tag)

    # CONNECTION
    def connect(self):
        r"""Establish the connection."""
        self.times_connected += 1
        parameters = pika.URLParameters(self.url)
        kwargs = dict(on_open_callback=self.on_connection_open,
                      on_open_error_callback=self.on_connection_open_error)
        if _pika_version_maj < 1:
            kwargs['stop_ioloop_on_close'] = False
        self.connection = pika.SelectConnection(parameters, **kwargs)

    def on_connection_open(self, connection):
        r"""Actions that must be taken when the connection is opened.
        Add the close connection callback and open the RabbitMQ channel."""
        self.debug('::Connection opened')
        connection.add_on_close_callback(self.on_connection_closed)
        self.open_channel()

    def on_connection_open_error(self, unused_connection):  # pragma: debug
        r"""Actions that must be taken when the connection fails to open."""
        self.debug('::Connection could not be opened')
        self.close()
        raise Exception('Could not connect.')

    def on_connection_closed(self, connection, *args):
        r"""Actions that must be taken when the connection is closed. Set the
        channel to None. If the connection is meant to be closing, stop the
        IO loop. Otherwise, wait 5 seconds and try to reconnect."""
        if _pika_version_maj < 1:
            reply_code = args[0]
            reply_text = args[1]
        # else:
        #     reply_code = 0
        #     reply_text = args[0]
        with self.rmq_lock:
            self.debug('::on_connection_closed code %s (code = %d)', reply_text,
                       reply_code)
            if self._closing or reply_code == 200:
                connection.ioloop.stop()
                self.connection = None
                self._closing = False
                self._qres_event.set()
            else:
                self.warning('Connection closed, reopening in %f seconds: (%s) %s',
                             self.sleeptime, reply_code, reply_text)
                self._reconnecting = True
                if _pika_version_maj < 1:
                    connection.add_timeout(self.sleeptime, self.reconnect)
                # else:
                #     connection.ioloop.call_later(self.sleeptime, self.reconnect)

    def reconnect(self):
        r"""Try to re-establish a connection and resume a new IO loop."""
        self.debug('')
        # This is the old connection IOLoop instance, stop its ioloop
        # with self.rmq_lock:
        self.connection.ioloop.stop()
        if not self._closing:
            # self.run_thread()
            # Create a new connection
            self.connect()
            # There is now a new connection, needs a new ioloop to run
            self._reconnecting = False
            self.connection.ioloop.start()

    # CHANNEL
    @property
    def channel_open(self):
        r"""bool: True if connection ready for messages, False otherwise."""
        with self.rmq_lock:
            if self.channel is None or self.connection is None:
                return False
            if self.channel.is_open:
                if not self.channel.is_closing:
                    return True
            return False  # pragma: debug

    @property
    def channel_stable(self):
        r"""bool: True if the connection ready for messages and not about to
        close. False otherwise."""
        with self.rmq_lock:
            return (self.channel_open and (not self._closing)
                    and (not self._opening))
        
    def open_channel(self):
        r"""Open a RabbitMQ channel."""
        self.debug('::Creating a new channel')
        self.connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        r"""Actions to perform after a channel is opened. Add the channel
        close callback and setup the exchange."""
        self.debug('::Channel opened')
        self.channel = channel
        channel.add_on_close_callback(self.on_channel_closed)
        self.setup_exchange(self.exchange)

    def on_channel_closed(self, channel, *args):
        r"""Actions to perform when the channel is closed. Close the
        connection."""
        with self.rmq_lock:
            if _pika_version_maj < 1:
                reply_code = args[0]
                reply_text = args[1]
                self.debug('::channel %i was closed: (%s) %s',
                           channel, reply_code, reply_text)
            # else:
            #     reason = args[0]
            #     self.debug('::channel %i was closed: %s',
            #                channel, reason)
            if not (channel.connection.is_closing or channel.connection.is_closed):
                channel.connection.close()
            self.channel = None
            # if self.connection is not None:
            #     self.connection.close()

    # EXCHANGE
    def setup_exchange(self, exchange_name):
        r"""Setup the exchange."""
        self.debug('::Declaring exchange %s', exchange_name)
        self.channel.exchange_declare(callback=self.on_exchange_declareok,
                                      exchange=exchange_name,
                                      auto_delete=True)

    def on_exchange_declareok(self, unused_frame):
        r"""Actions to perform once an exchange is succesfully declared.
        Set up the queue."""
        self.debug('::Exchange declared')
        self.setup_queue()

    # QUEUE
    def setup_queue(self):
        r"""Set up the message queue."""
        self.debug('::Declaring queue %s', self.queue)
        if self.direction == 'recv' and not self.queue:
            exclusive = False  # True
        else:
            exclusive = False
        if self.queue.startswith('amq.'):
            passive = True
        else:
            passive = False
        self.channel.queue_declare(queue=self.queue,
                                   callback=self.on_queue_declareok,
                                   exclusive=exclusive,
                                   # , auto_delete=True,
                                   passive=passive)

    def on_queue_declareok(self, method_frame):
        r"""Actions to perform once the queue is succesfully declared. Bind
        the queue."""
        self.debug('::Binding')
        with self.rmq_lock:
            if not self.queue:
                self.address += method_frame.method.queue
        self.channel.queue_bind(callback=self.on_bindok,
                                exchange=self.exchange,
                                # routing_key=self.routing_key,
                                queue=self.queue)
        
    def on_bindok(self, unused_frame):
        r"""Actions to perform once the queue is succesfully bound. Start
        consuming messages."""
        self.debug('::Queue bound')
        self.channel.basic_qos(prefetch_count=1)
        self.channel.add_on_cancel_callback(self.on_cancelok)
        if self.direction == 'recv':
            kwargs = dict(on_message_callback=self.on_message,
                          queue=self.queue)
            if _pika_version_maj < 1:
                kwargs['consumer_callback'] = kwargs.pop('on_message_callback')
            self.consumer_tag = self.channel.basic_consume(**kwargs)
        with self.rmq_lock:
            self._opening = False

    # GENERAL
    def on_cancelok(self, unused_frame):
        r"""Actions to perform after succesfully cancelling consumption. Closes
        the channel."""
        self.debug('::on_cancelok()')
        with self.rmq_lock:
            self.close_queue()
            self.close_channel()

    def close_connection(self):
        r"""Stop the ioloop and close the connection."""
        if self.connection:  # pragma: debug
            self.connection.ioloop.stop()
        super(RMQAsyncComm, self).close_connection()

    def force_close(self):  # pragma: debug
        r"""Force stop by removing the queue and stopping the IO loop."""
        with self.rmq_lock:
            self.close_queue()
            self.close_connection()
            self.channel = None
            self.connection = None
            self._closing = False

    def purge(self):
        r"""Remove all messages from the associated queue."""
        with self.rmq_lock:
            self._buffered_messages = []
            if not self.channel_stable:  # pragma: debug
                return
            super(RMQAsyncComm, self).purge()



import os
from yggdrasil import doctools
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType
from yggdrasil.metaschema.datatypes.tests.test_MetaschemaType import (
    TestMetaschemaType)
from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty
from yggdrasil.drivers.ModelDriver import ModelDriver
from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver
from yggdrasil.drivers.CompiledModelDriver import (
    CompiledModelDriver, CompilerBase, LinkerBase, ArchiverBase)

tables_dir = os.path.join(os.path.dirname(__file__), 'tables')
dir_list = [tables_dir]
class_list = [ModelDriver, InterpretedModelDriver, CompiledModelDriver,
              CompilerBase, LinkerBase, ArchiverBase,
              MetaschemaType, MetaschemaProperty]


for x in dir_list:
    if not os.path.isdir(x):
        os.mkdir(x)

        
doctools.write_component_table(fname_dir=tables_dir, verbose=False)
for c in class_list:
    doctools.write_classdocs_table(c, fname_dir=tables_dir, verbose=False)
doctools.write_datatype_table(fname_dir=tables_dir, verbose=False)
doctools.write_datatype_mapping_table(fname_dir=tables_dir, verbose=False)
doctools.write_comm_devnotes_table(fname_dir=tables_dir, verbose=False)



from yggdrasil.drivers.ConnectionDriver import ConnectionDriver


class InputDriver(ConnectionDriver):
    r"""Driver for receiving input from another model's comm via a local comm.

    Args:
        name (str): The name of the local message comm that the driver should
            connect to.
        args (str): The name of the other message comm that the driver should
            connect to.
        **kwargs: Additional keyword arguments are passed to parent class.

    """

    _connection_type = 'input'
    _direction = 'input'
    _schema_subtype_description = ('Connection between one or more comms/files '
                                   'and a model.')

    def __init__(self, name, args, **kwargs):
        kwargs.setdefault('icomm_kws', {})
        kwargs['icomm_kws']['name'] = args
        if kwargs['icomm_kws'].get('comm', self._icomm_type) == 'RMQComm':
            kwargs['icomm_kws']['queue'] = args
        super(InputDriver, self).__init__(name, **kwargs)



from yggdrasil.tests import assert_raises
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent


def valid_function():  # pragma: debug
    r"""Valid function for testing."""
    pass


class TestFunctionMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for FunctionMetaschemaType class with float."""

    _mod = 'FunctionMetaschemaType'
    _cls = 'FunctionMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestFunctionMetaschemaType, self).__init__(*args, **kwargs)
        self._value = valid_function
        self._valid_encoded = [self.typedef]
        self._valid_decoded = [self._value]
        self._invalid_encoded = [{}]
        self._invalid_decoded = [object]
        self._compatible_objects = [(self._value, self._value, None)]
        self._valid_normalize += [('%s:valid_function' % __name__, valid_function)]

    def test_decode_data_errors(self):
        r"""Test errors in decode_data."""
        assert_raises(ValueError, self.import_cls.decode_data, 'hello', None)
        assert_raises(AttributeError, self.import_cls.decode_data,
                      'yggdrasil:invalid', None)



from yggdrasil import tools
import yggdrasil.drivers.tests.test_ConnectionDriver as parent


class TestServerResponseParam(parent.TestConnectionParam):
    r"""Test parameters for ServerResponseDriver class."""

    def __init__(self, *args, **kwargs):
        super(TestServerResponseParam, self).__init__(*args, **kwargs)
        self.driver = 'ServerResponseDriver'
        self.args = None
        self.attr_list += ['comm', 'msg_id', 'model_response_name',
                           'model_response_address', 'response_address']
        self.comm_name = tools.get_default_comm()
        self.server_comm = tools.get_default_comm()
        self.icomm_name = self.comm_name
        self.ocomm_name = self.server_comm

    @property
    def inst_args(self):
        r"""tuple: Driver arguments."""
        out = super(TestServerResponseParam, self).inst_args
        out[0] = None  # Force driver to create an address
        return out
    
    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        out = super(TestServerResponseParam, self).inst_kwargs
        out['comm'] = self.server_comm
        return out


class TestServerResponseDriverNoStart(TestServerResponseParam,
                                      parent.TestConnectionDriverNoStart):
    r"""Test class for ServerResponseDriver class without start."""
    pass


class TestServerResponseDriver(TestServerResponseParam,
                               parent.TestConnectionDriver):
    r"""Test class for ServerResponseDriver class."""

    def test_send_recv(self):
        r"""Test sending/receiving small message."""
        super(TestServerResponseDriver, self).test_send_recv()
        assert(self.instance._used)
        assert(not self.instance.is_valid)

    def test_send_recv_nolimit(self):
        r"""Test sending/receiving large message."""
        super(TestServerResponseDriver, self).test_send_recv_nolimit()
        assert(self.instance._used)
        assert(not self.instance.is_valid)



import os
import copy
from yggdrasil import platform
from yggdrasil.drivers.CModelDriver import (
    CCompilerBase, CModelDriver, GCCCompiler, ClangCompiler)


class CPPCompilerBase(CCompilerBase):
    r"""Base class for C++ compilers."""
    languages = ['c++']
    default_executable_env = 'CXX'
    default_flags_env = 'CXXFLAGS'
    cpp_std = 'c++11'
    search_path_flags = ['-E', '-v', '-xc++', '/dev/null']
    default_linker = None
    default_executable = None

    @classmethod
    def get_flags(cls, **kwargs):
        r"""Get a list of compiler flags.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Compiler flags.

        """
        out = super(CCompilerBase, cls).get_flags(**kwargs)
        # Add standard library flag
        std_flag = None
        for i, a in enumerate(out):
            if a.startswith('-std='):
                std_flag = i
                break
        if std_flag is None:
            out.append('-std=%s' % cls.cpp_std)
        return out
    

class GPPCompiler(CPPCompilerBase, GCCCompiler):
    r"""Interface class for G++ compiler/linker."""
    toolname = 'g++'


class ClangPPCompiler(CPPCompilerBase, ClangCompiler):
    r"""clang++ compiler on Apple Mac OS."""
    toolname = 'clang++'


class CPPModelDriver(CModelDriver):
    r"""Class for running C++ models."""
                
    _schema_subtype_description = ('Model is written in C++.')
    language = 'c++'
    language_ext = ['.cpp', '.CPP', '.cxx', '.C', '.c++', '.cc', '.cp', '.tcc',
                    '.hpp', '.HPP', '.hxx', '.H', '.h++', '.hh', '.hp', '.h']
    language_aliases = ['cpp']
    base_languages = ['c']
    interface_library = 'ygg++'
    # To prevent inheritance
    default_compiler = None
    default_linker = None
    type_map = dict(
        CModelDriver.type_map,
        array='std::vector<void*>',
        object='std::map<char*,void*>',
        schema='MetaschemaType')
    function_param = dict(
        CModelDriver.function_param,
        exec_prefix=('#include <iostream>\n'
                     '#include <exception>\n'),
        # print='std::cout << "{message}" << std::endl;',
        error='throw \"{error_msg}\";',
        try_begin='try {',
        try_error_type='const std::exception&',
        try_except='}} catch ({error_type} {error_var}) {{')
    
    @staticmethod
    def after_registration(cls):
        r"""Operations that should be performed to modify class attributes after
        registration."""
        if cls.default_compiler is None:
            if platform._is_linux:
                cls.default_compiler = 'g++'
            elif platform._is_mac:
                cls.default_compiler = 'clang++'
            elif platform._is_win:  # pragma: windows
                cls.default_compiler = 'cl'
        cls.function_param['print'] = 'std::cout << "{message}" << std::endl;'
        CModelDriver.after_registration(cls)
        internal_libs = copy.deepcopy(cls.internal_libraries)
        internal_libs[cls.interface_library] = internal_libs.pop(
            CModelDriver.interface_library)
        internal_libs[cls.interface_library]['source'] = os.path.join(
            cls.get_language_dir(),
            os.path.splitext(os.path.basename(
                internal_libs[cls.interface_library]['source']))[0]
            + cls.language_ext[0])
        internal_libs[cls.interface_library]['include_dirs'].append(
            cls.get_language_dir())
        cls.internal_libraries = internal_libs

    @classmethod
    def write_try_except(cls, try_contents, except_contents, error_var='e',
                         error_type=None, **kwargs):
        r"""Return the lines required to complete a try/except block.

        Args:
            try_contents (list): Lines of code that should be executed inside
                the try block.
            except_contents (list): Lines of code that should be executed inside
                the except block.
            error_var (str, optional): Name of variable where the caught error
                should be stored. Defaults to 'e'. If '...', the catch clause
                will catch all errors, but there will not be a name error.
            error_type (str, optional): Name of error type that should be caught.
                If not provided, defaults to None and will be set based on the
                class function_param entry for 'try_error_type'. If '...', the
                catch clause will catch all errors and error_var will be
                ignored.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            Lines of code perfoming a try/except block.

        """
        if (error_type == '...') or (error_var == '...'):
            error_type = ''
            error_var = '...'
        kwargs.update(error_var=error_var, error_type=error_type)
        return super(CPPModelDriver, cls).write_try_except(
            try_contents, except_contents, **kwargs)



from yggdrasil.metaschema.properties.tests import (
    test_MetaschemaProperty as parent)


class ValidArgsClass1(object):
    def __init__(self, a, b):
        self._input_args = {'a': int(a), 'b': int(b)}
        

class ValidArgsClass2(ValidArgsClass1):
    def get_input_args(self):
        return self._input_args
    

class ValidArgsClass3(ValidArgsClass2):
    @property
    def input_arguments(self):
        return self._input_args


class TestArgsMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for ArgsMetaschemaProperty class."""
    
    _mod = 'ArgsMetaschemaProperty'
    _cls = 'ArgsMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestArgsMetaschemaProperty, self).__init__(*args, **kwargs)
        valid_type = {'a': {'type': 'int'}, 'b': {'type': 'int'}}
        invalid_type = {'a': {'type': 'int'}, 'b': {'type': 'float'}}
        self._valid = [(ValidArgsClass1(0, 1), valid_type),
                       (ValidArgsClass2(0, 1), valid_type),
                       (ValidArgsClass3(0, 1), valid_type)]
        self._invalid = [(ValidArgsClass1(0, 1), invalid_type),
                         (ValidArgsClass2(0, 1), invalid_type),
                         (ValidArgsClass3(0, 1), invalid_type)]
        self._encode_errors = [int(1)]
        self._valid_compare = [(valid_type, valid_type)]
        self._invalid_compare = [(valid_type, invalid_type)]



import unittest
import jsonschema
from yggdrasil.metaschema.datatypes import _jsonschema_ver_maj
from yggdrasil.metaschema import normalizer, get_metaschema, get_validator
from yggdrasil.tests import assert_equal, assert_raises


def test_create_normalizer():
    r"""Test create normalizer with default types."""
    cls = normalizer.create(get_metaschema())
    assert_equal(cls({'type': 'int'}).normalize('1'), '1')


def test_normalize_schema():
    r"""Test normalize_schema method on Normalizer."""
    kwargs = dict(normalizer_validators={'invalid': None})
    cls = normalizer.create(get_metaschema())
    cls.normalize_schema({'type': 'int'}, **kwargs)

    
@unittest.skipIf(_jsonschema_ver_maj >= 3, "JSON Schema >= 3.0.0")
def test_normalizer_iter_errors_js2():
    r"""Test normalizer iter_errors_js2 method."""
    cls = get_validator()
    s = {'type': 'array', 'items': [True]}
    x = cls(s)
    x.validate([True])
    s = {'type': 'array', 'items': [False]}
    x = cls(s)
    assert_raises(jsonschema.exceptions.ValidationError,
                  x.validate, [True])



from yggdrasil.interface import YggInput, YggOutput


def calc_photosynthesis_rate(T, CO2, light):
    r"""Calculate the rate of photosynthesis from environment properties.

    Args:
        T (float): Temperature.
        CO2 (float): CO2 concentration.
        light (float): Light intensity.

    Returns:
        float: Photosynthesis rate.

    """
    return light * CO2 / T


if __name__ == '__main__':
    in_temp = YggInput('temperature')
    in_co2 = YggInput('co2')
    in_light = YggInput('light_intensity')
    out_photo = YggOutput('photosynthesis_rate', '%lf\n')
    
    # Receive temperature & CO2 concentration of environment
    flag, msg = in_temp.recv()
    if not flag:
        raise RuntimeError("photosynthesis: Failed to receive temperature.")
    T = msg[0]
    print("photosynthesis: T = %f" % T)
    flag, msg = in_co2.recv()
    if not flag:
        raise RuntimeError("photosynthesis: Failed to receive CO2 concentration.")
    CO2 = msg[0]
    print("photosynthesis: CO2 = %f" % CO2)

    # Loop over light intensities
    while True:
        flag, msg = in_light.recv()
        if not flag:
            print("photosynthesis: No more input.")
            break
        LI = msg[0]
        PR = calc_photosynthesis_rate(T, CO2, LI)
        print("photosynthesis: light intensity = %f " % LI
              + "--> photosynthesis rate = %f" % PR)
        flag = out_photo.send(PR)
        if not flag:
            raise RuntimeError("photosynthesis: Failed to send output.")



#!/usr/bin/python
import os
import sys
import copy
import logging
import traceback


logger = logging.getLogger(__name__)


def yggrun():
    r"""Start a run."""
    from yggdrasil import runner
    prog = sys.argv[0].split(os.path.sep)[-1]
    # Print help
    if '-h' in sys.argv:
        print('Usage: yggrun [YAMLFILE1] [YAMLFILE2]...')
        return
    models = sys.argv[1:]
    yggRunner = runner.get_runner(models, ygg_debug_prefix=prog)
    try:
        yggRunner.run()
        yggRunner.debug("runner returns, exiting")
    except Exception as ex:
        yggRunner.pprint("yggrun exception: %s" % type(ex))
        print(traceback.format_exc())
    print('')


def yggcc():
    r"""Compile C/C++ program."""
    from yggdrasil.drivers import CModelDriver
    # prog = sys.argv[0].split(os.path.sep)[-1]
    src = sys.argv[1:]
    out = CModelDriver.CModelDriver.call_compile(src)
    print("executable: %s" % out)


def cc_flags():
    r"""Get the compiler flags necessary for including the interface
    library in a C or C++ program.

    Returns:
        list: The necessary compiler flags and preprocessor definitions.

    """
    from yggdrasil.drivers import CModelDriver
    print(' '.join(CModelDriver.CModelDriver.get_compiler_flags()))


def ld_flags():
    r"""Get the linker flags necessary for calling functions/classes from
    the interface library in a C or C++ program.

    Returns:
        list: The necessary library linking flags.

    """
    from yggdrasil.drivers import CModelDriver
    print(' '.join(CModelDriver.CModelDriver.get_linker_flags()))


def rebuild_c_api():
    r"""Rebuild the C/C++ API."""
    from yggdrasil.drivers import CModelDriver, CPPModelDriver
    if CModelDriver.CModelDriver.is_installed():
        CModelDriver.CModelDriver.compile_dependencies(overwrite=True)
        # TODO: Check that this compiles library correctly
        CPPModelDriver.CPPModelDriver.compile_dependencies(overwrite=True)
    else:
        raise Exception("The libraries necessary for running models written in "
                        "C/C++ could not be located.")

    
def regen_metaschema():
    r"""Regenerate the yggdrasil metaschema."""
    from yggdrasil import metaschema
    if os.path.isfile(metaschema._metaschema_fname):
        os.remove(metaschema._metaschema_fname)
    metaschema._metaschema = None
    metaschema._validator = None
    metaschema.get_metaschema()
    

def regen_schema():
    r"""Regenerate the yggdrasil schema."""
    from yggdrasil import schema
    if os.path.isfile(schema._schema_fname):
        os.remove(schema._schema_fname)
    schema.clear_schema()
    schema.init_schema()


def validate_yaml():
    r"""Validate a set of or or more YAMLs defining an integration."""
    from yggdrasil import yamlfile
    files = sys.argv[1:]
    yamlfile.parse_yaml(files)
    logger.info("Validation succesful.")


def update_config():
    r"""Update the user config file for yggdrasil."""
    from yggdrasil import config, tools
    from yggdrasil.components import import_component
    overwrite = ('--overwrite' in sys.argv)
    drv = [import_component('model', l) for l in tools.get_supported_lang()]
    config.update_language_config(drv, overwrite=overwrite,
                                  verbose=True)


def yggtime_comm():
    r"""Plot timing statistics comparing the different communication mechanisms."""
    from yggdrasil import timing
    timing.plot_scalings(compare='commtype')


def yggtime_lang():
    r"""Plot timing statistics comparing the different languages."""
    from yggdrasil import timing
    timing.plot_scalings(compare='language')


def yggtime_os():
    r"""Plot timing statistics comparing the different operating systems."""
    from yggdrasil import timing
    timing.plot_scalings(compare='platform')


def yggtime_py():
    r"""Plot timing statistics comparing the different versions of Python."""
    from yggdrasil import timing
    timing.plot_scalings(compare='python')


def yggtime_paper():
    r"""Create plots for timing."""
    from yggdrasil import timing
    _lang_list = timing._lang_list
    _lang_list_nomatlab = copy.deepcopy(_lang_list)
    _lang_list_nomatlab.remove('matlab')
    timing.plot_scalings(compare='platform', python_ver='2.7')
    # All plots on Linux, no matlab
    timing.plot_scalings(compare='comm_type', platform='Linux', python_ver='2.7')
    timing.plot_scalings(compare='python_ver', platform='Linux')
    timing.plot_scalings(compare='language', platform='Linux', python_ver='2.7',
                         compare_values=_lang_list_nomatlab)
    # Language comparision on MacOS, with matlab
    timing.plot_scalings(compare='language', platform='MacOS', python_ver='2.7',
                         compare_values=_lang_list)


if __name__ == '__main__':
    yggrun()
    sys.exit(0)



import os
import copy
import pprint
import pystache
import yaml
from yggdrasil import backwards
from yggdrasil.schema import standardize, get_schema


def load_yaml(fname):
    r"""Parse a yaml file defining a run.

    Args:
        fname (str, file, dict): Path to a YAML file, an open file descriptor
            to a file containing a YAML, or a loaded YAML document.

    Returns:
        dict: Contents of yaml file.

    """
    opened = False
    if isinstance(fname, dict):
        yamlparsed = copy.deepcopy(fname)
        yamlparsed.setdefault('working_dir', os.getcwd())
        return yamlparsed
    elif isinstance(fname, str):
        fname = os.path.realpath(fname)
        if not os.path.isfile(fname):
            raise IOError("Unable locate yaml file %s" % fname)
        fd = open(fname, 'r')
        opened = True
    else:
        fd = fname
        if (hasattr(fd, 'name') and (not fd.name.startswith('<'))):
            fname = fd.name
        else:
            fname = os.path.join(os.getcwd(), 'stream')
    # Mustache replace vars
    yamlparsed = fd.read()
    yamlparsed = pystache.render(
        backwards.StringIO(yamlparsed).getvalue(), dict(os.environ))
    yamlparsed = yaml.safe_load(yamlparsed)
    if not isinstance(yamlparsed, dict):  # pragma: debug
        raise ValueError("Loaded yaml is not a dictionary.")
    yamlparsed['working_dir'] = os.path.dirname(fname)
    if opened:
        fd.close()
    return yamlparsed


def prep_yaml(files):
    r"""Prepare yaml to be parsed by jsonschema including covering backwards
    compatible options.

    Args:
        files (str, list): Either the path to a single yaml file or a list of
            yaml files. Entries can also be opened file descriptors for files
            containing YAML documents or pre-loaded YAML documents.

    Returns:
        dict: YAML ready to be parsed using schema.

    """
    # Load each file
    if not isinstance(files, list):
        files = [files]
    yamls = [load_yaml(f) for f in files]
    # Standardize format of models and connections to be lists and
    # add working_dir to each
    comp_keys = ['models', 'connections']
    for yml in yamls:
        standardize(yml, comp_keys)
        for k in comp_keys:
            for x in yml[k]:
                if isinstance(x, dict):
                    x.setdefault('working_dir', yml['working_dir'])
    # Combine models & connections
    yml_all = {}
    for k in comp_keys:
        yml_all[k] = []
        for yml in yamls:
            yml_all[k] += yml[k]
    return yml_all


def parse_yaml(files):
    r"""Parse list of yaml files.

    Args:
        files (str, list): Either the path to a single yaml file or a list of
            yaml files.

    Raises:
        ValueError: If the yml dictionary is missing a required keyword or has
            an invalid value.
        RuntimeError: If one of the I/O channels is not initialized with driver
            information.

    Returns:
        dict: Dictionary of information parsed from the yamls.

    """
    s = get_schema()
    # Parse files using schema
    yml_prep = prep_yaml(files)
    # print('prepped')
    # pprint.pprint(yml_prep)
    yml_norm = s.validate(yml_prep, normalize=True)
    # print('normalized')
    # pprint.pprint(yml_norm)
    # Parse models, then connections to ensure connections can be processed
    existing = None
    for k in ['models', 'connections']:
        for yml in yml_norm[k]:
            existing = parse_component(yml, k[:-1], existing=existing)
    # Make sure that I/O channels initialized
    for io in ['input', 'output']:
        for k, v in existing[io].items():
            if 'driver' not in v:
                raise RuntimeError("No driver established for %s channel %s" % (
                    io, k))
    # Link io drivers back to models
    existing = link_model_io(existing)
    # print('drivers')
    # pprint.pprint(existing)
    return existing


def parse_component(yml, ctype, existing=None):
    r"""Parse a yaml entry for a component, adding it to the list of
    existing components.

    Args:
        yml (dict): YAML dictionary for a component.
        ctype (str): Component type. This can be 'input', 'output',
            'model', or 'connection'.
        existing (dict, optional): Dictionary of existing components.
            Defaults to empty dict.

    Raises:
        TypeError: If yml is not a dictionary.
        ValueError: If dtype is not 'input', 'output', 'model', or
            'connection'.
        ValueError: If the component already exists.

    Returns:
        dict: All components identified.

    """
    s = get_schema()
    if not isinstance(yml, dict):
        raise TypeError("Component entry in yml must be a dictionary.")
    ctype_list = ['input', 'output', 'model', 'connection',
                  'model_input', 'model_output']
    if existing is None:
        existing = {k: {} for k in ctype_list}
    if ctype not in ctype_list:
        raise ValueError("'%s' is not a recognized component.")
    # Parse based on type
    if ctype == 'model':
        existing = parse_model(yml, existing)
    elif ctype == 'connection':
        existing = parse_connection(yml, existing)
    elif ctype in ['input', 'output']:
        for k in ['icomm_kws', 'ocomm_kws']:
            if k not in yml:
                continue
            for x in yml[k]['comm']:
                if 'comm' not in x:
                    if 'filetype' in x:
                        x['comm'] = s['file'].subtype2class[x['filetype']]
                    elif 'commtype' in x:
                        x['comm'] = s['comm'].subtype2class[x['commtype']]
    # Ensure component dosn't already exist
    if yml['name'] in existing[ctype]:
        pprint.pprint(existing)
        pprint.pprint(yml)
        raise ValueError("%s is already a registered '%s' component." % (
            yml['name'], ctype))
    existing[ctype][yml['name']] = yml
    return existing


def parse_model(yml, existing):
    r"""Parse a yaml entry for a model.

    Args:
        yml (dict): YAML dictionary for a model.
        existing (dict): Dictionary of existing components.

    Returns:
        dict: Updated log of all entries.

    """
    _lang2driver = get_schema()['model'].subtype2class
    yml['driver'] = _lang2driver[yml.pop('language')]
    # Add server driver
    if yml.get('is_server', False):
        srv = {'name': yml['name'],
               'commtype': 'default',
               'datatype': {'type': 'bytes'},
               'driver': 'ServerDriver',
               'args': yml['name'] + '_SERVER',
               'working_dir': yml['working_dir']}
        yml['inputs'].append(srv)
        yml['clients'] = []
    # Add client driver
    if yml.get('client_of', []):
        srv_names = yml['client_of']
        yml['client_of'] = srv_names
        for srv in srv_names:
            cli = {'name': '%s_%s' % (srv, yml['name']),
                   'commtype': 'default',
                   'datatype': {'type': 'bytes'},
                   'driver': 'ClientDriver',
                   'args': srv + '_SERVER',
                   'working_dir': yml['working_dir']}
            yml['outputs'].append(cli)
    # Model index and I/O channels
    yml['model_index'] = len(existing['model'])
    for io in ['inputs', 'outputs']:
        for x in yml[io]:
            x['model_driver'] = [yml['name']]
            existing = parse_component(x, io[:-1], existing=existing)
    return existing
            
    
def parse_connection(yml, existing):
    r"""Parse a yaml entry for a connection between I/O channels.

    Args:
        yml (dict): YAML dictionary for a connection.
        existing (dict): Dictionary of existing components.

    Raises:
        RuntimeError: If the 'inputs' entry is not a model output or file.
        RuntimeError: If neither the 'inputs' or 'outputs' entries correspond
            to model I/O channels.

    Returns:
        dict: Updated log of all entries.

    """
    schema = get_schema()
    # File input
    is_file = {'inputs': [], 'outputs': []}
    iname_list = []
    for x in yml['inputs']:
        is_file['inputs'].append(schema.is_valid_component('file', x))
        if is_file['inputs'][-1]:
            fname = os.path.expanduser(x['name'])
            if not os.path.isabs(fname):
                fname = os.path.join(x['working_dir'], fname)
            fname = os.path.normpath(fname)
            if (not os.path.isfile(fname)) and (not x.get('wait_for_creation', False)):
                raise RuntimeError(("Input '%s' not found in any of the registered "
                                    + "model outputs and is not a file.") % x['name'])
            x['address'] = fname
        else:
            iname_list.append(x['name'])
    # File output
    oname_list = []
    for x in yml['outputs']:
        is_file['outputs'].append(schema.is_valid_component('file', x))
        if is_file['outputs'][-1]:
            fname = os.path.expanduser(x['name'])
            if not x.get('in_temp', False):
                if not os.path.isabs(fname):
                    fname = os.path.join(x['working_dir'], fname)
                fname = os.path.normpath(fname)
            x['address'] = fname
        else:
            oname_list.append(x['name'])
    iname = ','.join(iname_list)
    oname = ','.join(oname_list)
    if not iname:
        args = oname
    elif not oname:
        args = iname
    else:
        args = '%s_to_%s' % (iname, oname)
    name = args
    # TODO: Use RMQ drivers when models are on different machines
    # ocomm_pair = ('default', 'rmq')
    # icomm_pair = ('rmq', 'default')
    # Output driver
    xo = None
    if iname:  # empty name results when all of the inputs are files
        iyml = yml['inputs']
        xo = {'name': iname, 'model_driver': [],
              'icomm_kws': {'comm': []},
              'ocomm_kws': {'comm': []}}
        for i, y in enumerate(iyml):
            if not is_file['inputs'][i]:
                xo['icomm_kws']['comm'].append(existing['output'][y['name']])
                xo['icomm_kws']['comm'][-1].update(**y)
                xo['model_driver'] += existing['output'][y['name']]['model_driver']
                del existing['output'][y['name']]
        # Add single non-file intermediate output comm if there are any non-file
        # outputs and an output comm for each file output
        if (sum(is_file['outputs']) < len(is_file['outputs'])):
            xo['ocomm_kws']['comm'].append({'name': args, 'no_suffix': True})
        for i, y in enumerate(yml['outputs']):
            if is_file['outputs'][i]:
                xo['ocomm_kws']['comm'].append(y)
        existing = parse_component(xo, 'output', existing)
        xo['args'] = args
        xo['driver'] = 'OutputDriver'
    # Input driver
    xi = None
    if oname:  # empty name results when all of the outputs are files
        oyml = yml['outputs']
        xi = {'name': oname, 'model_driver': [],
              'icomm_kws': {'comm': []},
              'ocomm_kws': {'comm': []}}
        for i, y in enumerate(oyml):
            if not is_file['outputs'][i]:
                xi['ocomm_kws']['comm'].append(existing['input'][y['name']])
                xi['ocomm_kws']['comm'][-1].update(**y)
                xi['model_driver'] += existing['input'][y['name']]['model_driver']
                del existing['input'][y['name']]
        # Add single non-file intermediate input comm if there are any non-file
        # inputs and an input comm for each file input
        if (sum(is_file['inputs']) < len(is_file['inputs'])):
            xi['icomm_kws']['comm'].append({'name': args, 'no_suffix': True})
        for i, y in enumerate(yml['inputs']):
            if is_file['inputs'][i]:
                xi['icomm_kws']['comm'].append(y)
        existing = parse_component(xi, 'input', existing)
        xi['args'] = args
        xi['driver'] = 'InputDriver'

    # Parse drivers
    
    # Transfer connection keywords to one connection driver
    conn_keys_gen = ['inputs', 'outputs']
    conn_keys = list(set(schema['connection'].properties) - set(conn_keys_gen))
    yml_conn = {}
    yml_conn.pop('name', None)
    for k in conn_keys:
        if k in yml:
            yml_conn[k] = yml[k]
    if xi is None:
        xo.update(**yml_conn)
    else:
        xi.update(**yml_conn)
    yml['name'] = name
    return existing


def link_model_io(existing):
    r"""Link I/O drivers back to the models they communicate with.

    Args:
        existing (dict): Dictionary of existing components.

    Returns:
        dict: Dictionary with I/O drivers added to models.

    """
    # Add fields
    for m in existing['model'].keys():
        existing['model'][m]['input_drivers'] = []
        existing['model'][m]['output_drivers'] = []
    # Add input dirvers
    for io in existing['input'].values():
        for m in io['model_driver']:
            existing['model'][m]['input_drivers'].append(io)
    # Add output dirvers
    for io in existing['output'].values():
        for m in io['model_driver']:
            existing['model'][m]['output_drivers'].append(io)
    return existing
